"""Identity module.

This module contains all the classes and method to handle the identity of an Asus device."""

from __future__ import annotations

import logging
from dataclasses import dataclass
from itertools import chain
from typing import Any, Awaitable, Callable, Optional, Tuple

from asusrouter.error import AsusRouterIdentityError
from asusrouter.modules.aimesh import AiMeshDevice
from asusrouter.modules.color import color_zone
from asusrouter.modules.data import AsusData
from asusrouter.modules.endpoint import (
    Endpoint,
    EndpointNoCheck,
    EndpointTools,
    EndpointType,
    check_available,
)
from asusrouter.modules.endpoint.onboarding import (
    process as process_onboarding,
)
from asusrouter.modules.endpoint.onboarding import read as read_onboarding
from asusrouter.modules.firmware import Firmware
from asusrouter.modules.wlan import WLAN_TYPE, Wlan
from asusrouter.tools import writers
from asusrouter.tools.cleaners import clean_dict
from asusrouter.tools.converters import (
    safe_exists,
    safe_list_from_string,
    safe_unpack_keys,
)

_LOGGER = logging.getLogger(__name__)

MAP_IDENTITY: Tuple = (
    ("serial_no", "serial"),
    ("label_mac", "mac"),
    ("lan_hwaddr", "lan_mac"),
    ("wan_hwaddr", "wan_mac"),
    ("productid", "product_id"),
    ("productid", "model"),
    ("firmver", "fw_major"),
    ("buildno", "fw_minor"),
    ("extendno", "fw_build"),
    ("rc_support", "services", safe_list_from_string),
    ("ss_support", "services", safe_list_from_string),
    ("led_val", "led", safe_exists),
    ("ledg_rgb1", "aura", safe_exists),
    ("ledg_rgb2", "aura_zone", color_zone),
)


@dataclass
class AsusDevice:  # pylint: disable=too-many-instance-attributes
    """Asus device class.

    This class contains information about an Asus device
    and can represent a router, AiMesh nore or range extender."""

    # Device-defining values
    serial: Optional[str] = None
    mac: Optional[str] = None
    product_id: Optional[str] = None
    model: Optional[str] = None
    brand: str = "ASUSTek"

    # Supported features
    aimesh: bool = False

    # Device information
    firmware: Optional[Firmware] = None
    merlin: bool = False
    wlan: Optional[list[Wlan]] = None
    endpoints: Optional[dict[EndpointType, bool]] = None
    services: Optional[list[str]] = None

    # Flags for device features
    aura: bool = False
    aura_zone: int = 0
    dsl: bool = False
    led: bool = False
    ookla: bool = False
    vpn_status: bool = False


async def collect_identity(
    api_hook: Callable[..., Awaitable[dict[str, Any]]],
    api_query: Callable[..., Awaitable[Any]],
) -> AsusDevice:
    """Collect device identity."""

    _LOGGER.debug("Collecting device identity")

    # Prepare a request
    request_values = []
    for map_item in MAP_IDENTITY:
        key, _, _ = safe_unpack_keys(map_item)
        request_values.append(key)
    request = writers.nvram(request_values)

    # Get the identity data
    try:
        identity_map = await api_hook(request)
    except Exception as ex:  # pylint: disable=broad-except
        raise AsusRouterIdentityError(
            "Failed to collect identity data from the router"
        ) from ex
    _LOGGER.debug("Identity collected")

    # Read the identity
    identity = _read_nvram(identity_map)
    _LOGGER.debug("Identity read")

    # Process services
    identity["aimesh"] = "amas" in identity["services"]

    # Check endpoints
    endpoints, onboarding = await _check_endpoints(api_query)
    identity["endpoints"] = endpoints
    _LOGGER.debug("Endpoints checked")
    # Manually assign Aura endpoint value
    identity["endpoints"][EndpointTools.AURA] = identity.get("aura", False)

    # Check onboarding to get nice model name
    this_device = onboarding.get(identity["mac"])
    if isinstance(this_device, AiMeshDevice):
        identity["model"] = this_device.model

    # Check if DSL
    if endpoints[Endpoint.DSL] is True:
        identity["dsl"] = True
        _LOGGER.debug("DSL-compatible device detected")

    # Check if Merlin
    if endpoints[Endpoint.SYSINFO] is True:
        identity["merlin"] = True
        _LOGGER.debug("Merlin FW detected")

    # Return the identity convered from a dict
    return AsusDevice(**identity)


# UPDATE THIS METHOD
def _read_nvram(data: dict[str, Any]) -> dict[str, Any]:
    """Read the NVRAM identity data."""

    # Check the input data
    if not data:
        raise AsusRouterIdentityError("No nvram data received")

    # Create the identity dictionary
    identity: dict[str, Any] = {}

    # Loop through the identity map
    for map_item in MAP_IDENTITY:
        key, key_to_use, method = safe_unpack_keys(map_item)
        try:
            value = method(data[key]) if method else data[key]
            if key_to_use in identity:
                if isinstance(identity[key_to_use], list):
                    identity[key_to_use].extend(value)
                else:
                    identity[key_to_use] = value
            else:
                identity[key_to_use] = value
        except Exception as ex:
            raise ex

    # Clean up the identity dictionary
    identity = clean_dict(identity)

    # MAC (for some Merlin firmwares missing label_mac)
    if not identity.get("mac"):
        if identity.get("lan_mac"):
            identity["mac"] = identity["lan_mac"]
        elif identity.get("wan_mac"):
            identity["mac"] = identity["wan_mac"]
    # Remove extra MACs
    identity.pop("lan_mac")
    identity.pop("wan_mac")

    # Firmware
    # TODO: Optimize this
    firmware = Firmware(
        f"{identity['fw_major']}.{identity['fw_minor']}.{identity['fw_build']}"
    )
    identity["firmware"] = firmware
    identity.pop("fw_major")
    identity.pop("fw_minor")
    identity.pop("fw_build")

    # WLAN list
    identity["wlan"] = []
    for value in identity["services"]:
        if value in WLAN_TYPE:
            identity["wlan"].append(WLAN_TYPE[value])

    # OOKLA Speedtest
    if "ookla" in identity["services"]:
        identity["ookla"] = True

    return identity


async def _check_endpoints(
    api_hook: Callable[..., Awaitable[Any]],
) -> tuple[dict[EndpointType, bool], dict[str, Any]]:
    """Check which endpoints are available."""

    endpoints: dict[EndpointType, bool] = {}
    contents: dict[EndpointType, Any] = {}

    for endpoint in chain(Endpoint, EndpointTools):
        if endpoint.name in EndpointNoCheck.__members__:
            continue
        result, content = await check_available(endpoint, api_hook)
        endpoints[endpoint] = result
        contents[endpoint] = content

    onboarding = {}
    if endpoints[Endpoint.ONBOARDING] is True:
        onboarding = process_onboarding(
            read_onboarding(contents[Endpoint.ONBOARDING])
        ).get(AsusData.AIMESH, {})

    return (endpoints, onboarding)



"""Constants for modules."""

from __future__ import annotations

from typing import Any, Callable

# A converter type
MapConverterType = Callable[..., Any] | list[Callable[..., Any]]

# A map value type - this value should be found and converted
MapValueType = str | tuple[str] | tuple[str, MapConverterType]

# A map replace type - this value should be found, renamed and converted
MapReplaceType = str | tuple[str] | tuple[str, str] | tuple[str, str, MapConverterType]



"""Firmware release notes endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.tools.converters import clean_string


def read(content: str) -> dict[str, Any]:
    """Read firmware release note data."""

    raw_note = content.replace("\ufeff", "")
    raw_note = raw_note.replace("\uff1a", ":")  # Replace the full-width colon

    # Empty release note -> fast return
    if raw_note == "\n\n":
        return {}

    # Get only the changes
    # 1. Split the data into lines
    lines = raw_note.splitlines()
    # 2. Clean the lines
    clean_lines = []
    for line in lines:
        if "Firmware version" in line or "Release Note" in line:
            continue
        clean_line = clean_string(line)
        if clean_line:
            clean_lines.append(clean_line)
    # 3. Combine the lines into a single string again
    note = "\n".join(clean_lines)

    release_note: dict[str, Any] = {
        "release_note": note,
        "release_note_raw": raw_note,
    }

    return release_note


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process firmware release note data."""

    state: dict[AsusData, Any] = {
        AsusData.FIRMWARE_NOTE: data,
    }

    return state



"""Hook endpoint module."""

from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import Any, Optional, Tuple

from asusrouter.modules.aura import process_aura
from asusrouter.modules.connection import ConnectionState, ConnectionStatus
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.endpoint import data_get
from asusrouter.modules.endpoint.error import AccessError
from asusrouter.modules.led import AsusLED
from asusrouter.modules.parental_control import (
    KEY_PC_BLOCK_ALL,
    KEY_PC_STATE,
    AsusBlockAll,
    AsusParentalControl,
    read_pc_rules,
)
from asusrouter.modules.port_forwarding import (
    KEY_PORT_FORWARDING_LIST,
    KEY_PORT_FORWARDING_STATE,
    AsusPortForwarding,
    PortForwardingRule,
)
from asusrouter.modules.vpnc import AsusVPNC, AsusVPNType
from asusrouter.modules.wlan import MAP_GWLAN, MAP_WLAN, Wlan
from asusrouter.tools.converters import (
    run_method,
    safe_bool,
    safe_datetime,
    safe_int,
    safe_return,
    safe_speed,
    safe_unpack_key,
    safe_unpack_keys,
    safe_usage,
    safe_usage_historic,
)
from asusrouter.tools.readers import merge_dicts
from asusrouter.tools.readers import read_json_content as read  # noqa: F401

from .hook_const import (
    MAP_NETWORK,
    MAP_OVPN_SERVER_388,
    MAP_SPEEDTEST,
    MAP_VPNC_WIREGUARD,
    MAP_WAN,
    MAP_WAN_ITEM,
    MAP_WAN_ITEM_X,
    MAP_WIREGUARD_CLIENT,
    MAP_WIREGUARD_SERVER,
)

REQUIRE_HISTORY = True
REQUIRE_WLAN = True

_LOGGER = logging.getLogger(__name__)


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process hook data."""

    # For this endpoint, the received data always depends on the sent request.
    # So, we need to check which data is available and process it accordingly.
    # Otherwise, we can accidentally overwrite the data with empty values.

    state: dict[AsusData, Any] = {}

    # Get the passed awrguments
    history: dict[AsusData, AsusDataState] = data_get(data, "history") or {}
    wlan = data_get(data, "wlan") or []

    # Aura
    if "ledg_scheme" in data:
        state[AsusData.AURA] = process_aura(data)

    # CPU
    if "cpu_usage" in data:
        cpu_usage = data.get("cpu_usage", {})
        prev_cpu: Optional[AsusDataState] = history.get(AsusData.CPU)
        state[AsusData.CPU] = (
            process_cpu(cpu_usage, prev_cpu)
            if cpu_usage
            else prev_cpu.data
            if isinstance(prev_cpu, AsusDataState)
            else {}
        )

    # GWLAN
    if (
        "wl0.1_wpa_psk" in data
        or "wl1.1_wpa_psk" in data
        or "wl2.1_wpa_psk" in data
        or "wl3.1_wpa_psk" in data
    ):
        state[AsusData.GWLAN] = process_gwlan(data, wlan)

    # LED
    if "led_val" in data:
        state[AsusData.LED] = {
            "state": AsusLED(safe_int(data.get("led_val"), default=-999))
        }

    # Network
    if "netdev" in data:
        prev_network: Optional[AsusDataState] = history.get(AsusData.NETWORK)
        netdev = data.get("netdev")
        state[AsusData.NETWORK] = (
            process_network(netdev, prev_network)
            if netdev
            else prev_network.data
            if isinstance(prev_network, AsusDataState)
            else {}
        )

    # OpenVPN Server
    if "vpn_serverx_clientlist" in data:
        state[AsusData.OPENVPN_SERVER] = process_openvpn_server(data)

    # Parental control
    if KEY_PC_STATE in data:
        state[AsusData.PARENTAL_CONTROL] = process_parental_control(data)

    # Port forwarding
    if KEY_PORT_FORWARDING_STATE in data:
        state[AsusData.PORT_FORWARDING] = process_port_forwarding(data)

    # RAM
    if "memory_usage" in data:
        memory_usage = data.get("memory_usage", {})
        state[AsusData.RAM] = process_ram(memory_usage) if memory_usage else {}

    # Speedtest
    if "ookla_state" in data:
        speedtest = process_speedtest(data)
        state[AsusData.SPEEDTEST_RESULT] = speedtest.get("result")
        state[AsusData.SPEEDTEST] = speedtest.get("data")

    # VPNC
    if "vpnc_clientlist" in data:
        vpnc, vpnc_clientlist = process_vpnc(data)
        state[AsusData.OPENVPN_CLIENT] = vpnc[AsusVPNType.OPENVPN]
        state[AsusData.VPNC] = vpnc
        state[AsusData.VPNC_CLIENTLIST] = vpnc_clientlist
        state[AsusData.WIREGUARD_CLIENT] = vpnc[AsusVPNType.WIREGUARD]

    # WAN
    if "get_wan_unit" in data:
        state[AsusData.WAN] = process_wan(data)

    # WireGuard
    if "get_wgsc_status" in data:
        state[AsusData.WIREGUARD_SERVER] = process_wireguard_server(data)

    # WLAN
    if (
        "wl0_wpa_psk" in data
        or "wl1_wpa_psk" in data
        or "wl2_wpa_psk" in data
        or "wl3_wpa_psk" in data
    ):
        state[AsusData.WLAN] = process_wlan(data, wlan)
    
    # DSL
    if (
        "dsllog_dataratedown" in data
        or "dsllog_datarateup" in data
    ):
        state[AsusData.DSL] = process_dsl(data)

    return state


def process_cpu(
    cpu_usage: dict[str, Any], history: Optional[AsusDataState]
) -> dict[str | int, Any]:
    """Process CPU data."""

    cpu = process_cpu_usage(cpu_usage)

    # Get the previous data
    prev_cpu = history.data if history else None

    # Safe calculate of actual usage if previous data is available
    if prev_cpu:
        for item, after in cpu.items():
            if item in prev_cpu:
                before = prev_cpu[item]
                after["usage"] = safe_usage_historic(
                    after["used"],
                    after["total"],
                    before["used"],
                    before["total"],
                )

    return cpu


def process_cpu_usage(raw: dict[str, Any]) -> dict[str | int, Any]:
    """Process CPU usage."""

    # Populate total
    cpu: dict[str | int, Any] = {"total": {"total": 0.0, "used": 0.0}}

    # Process each core
    core = 1
    while f"cpu{core}_total" in raw:
        cpu[core] = {
            "total": int(raw[f"cpu{core}_total"]),
            "used": int(raw[f"cpu{core}_usage"]),
        }
        # Update the total
        cpu["total"]["total"] += cpu[core]["total"]
        cpu["total"]["used"] += cpu[core]["used"]

        core += 1

    return cpu


def process_gwlan(
    data: dict[str, Any], wlan_list: list[Wlan]
) -> dict[str, Any]:
    """Process GWLAN data."""

    gwlan = {}

    for interface in wlan_list:
        index = list(Wlan).index(interface)
        for gid in range(1, 4):
            info = {}
            for pair in MAP_GWLAN:
                key, method = safe_unpack_key(pair)
                info[key.format(f"{index}.{gid}")[6:]] = (
                    method(data.get(key.format(f"{index}.{gid}")))
                    if method
                    else data.get(key.format(f"{index}.{gid}"))
                )
            gwlan[f"{interface.value}_{gid}"] = info

    return gwlan


def process_network(
    netdev: dict[str, Any], history: Optional[AsusDataState]
) -> dict[str, Any]:
    """Process network data."""

    # Calculate RX and TX from the HEX values.
    network = process_network_usage(netdev)

    # Calculate speeds if previous data is available
    if history and history.data:
        time_delta = (
            datetime.now(timezone.utc) - history.timestamp
        ).total_seconds()
        network = process_network_speed(network, history.data, time_delta)

    return network


def process_network_speed(
    network: dict[str, dict[str, float]],
    prev_network: dict[str, dict[str, float]],
    time_delta: float | None,
) -> dict[str, dict[str, float]]:
    """Calculate network speed for a set period of time."""

    # Check values one by one
    for interface in network:
        # Skip if there is no previous data
        if interface not in prev_network:
            continue

        # Dictionary with speed values
        interface_speed = {}

        # Calculate speed for each traffic type
        for traffic_type, traffic_value in network[interface].items():
            prev_traffic_value = prev_network[interface].get(traffic_type)
            # Calculate speed only if previous value is available
            if prev_traffic_value:
                interface_speed[f"{traffic_type}_speed"] = 8 * safe_speed(
                    traffic_value,
                    prev_traffic_value,
                    time_delta,
                )
                continue
            # Otherwise, set speed to 0
            interface_speed[f"{traffic_type}_speed"] = 0.0

        # Update interface with speed values
        network[interface].update(interface_speed)

    return network


def process_network_usage(raw: dict[str, Any]) -> dict[str, Any]:
    """Process network usage data."""

    network = {}
    for key, key_to_use in MAP_NETWORK.items():
        data = {}
        for traffic_type in ("rx", "tx"):
            # Convert string with HEX value to int
            value = safe_int(raw.get(f"{key}_{traffic_type}"), base=16)
            # Check that value is integer
            if isinstance(value, int):
                data[traffic_type] = value
        if len(data) > 0:
            network[key_to_use] = data

    return network


def process_openvpn_server(data: dict[str, Any]) -> dict[int, Any]:
    """Process OpenVPN server data."""

    server = {}

    # Server data
    for keys in MAP_OVPN_SERVER_388:
        key, key_to_use, method = safe_unpack_keys(keys)
        state_value = data.get(key)
        if state_value:
            server[key_to_use] = run_method(state_value, method)

    # Clients
    clients = server.get("clients", "")
    clients = clients.replace("&#62", ">").replace("&#60", "<")
    clients = clients[1:-1].split("><")
    server["clients"] = clients

    return {1: server}


def process_parental_control(data: dict[str, Any]) -> dict[str, Any]:
    """Process parental control data."""

    parental_control: dict[str, Any] = {}

    # State
    parental_control["state"] = AsusParentalControl(
        safe_int(data.get(KEY_PC_STATE), default=-999)
    )

    # Block all
    parental_control["block_all"] = AsusBlockAll(
        safe_int(data.get(KEY_PC_BLOCK_ALL), default=-999)
    )

    # Rules
    parental_control["rules"] = read_pc_rules(data)

    return parental_control


def process_port_forwarding(data: dict[str, Any]) -> dict[str, Any]:
    """Process port forwarding data."""

    port_forwarding = {}

    # State
    port_forwarding["state"] = AsusPortForwarding(
        safe_int(data.get(KEY_PORT_FORWARDING_STATE), default=-999)
    )

    # Rules
    pf_list = data.get(KEY_PORT_FORWARDING_LIST)
    if pf_list:
        rules = []
        rule_list = pf_list.split("&#60")
        for rule in rule_list:
            if rule == str():
                continue
            part = rule.split("&#62")
            rules.append(
                PortForwardingRule(
                    name=safe_return(part[0]),
                    ip_address=part[2],
                    port=safe_return(part[3]),
                    protocol=part[4],
                    ip_external=safe_return(part[5]),
                    port_external=part[1],
                )
            )
        port_forwarding["rules"] = rules.copy()

    return port_forwarding


def process_ram(memory_usage: dict[str, Any]) -> dict[str, Any]:
    """Process RAM data."""

    ram: dict[str, Any] = {}
    # Data is in KiB. To get MB as they are shown in the device Web-GUI,
    # should be devided by 1024 (yes, those will be MiB)

    # Populate RAM with known values
    ram = {
        "free": safe_int(memory_usage.get("mem_free")),
        "total": safe_int(memory_usage.get("mem_total")),
        "used": safe_int(memory_usage.get("mem_used")),
    }
    # Calculate usage in percents
    if "used" in ram and "total" in ram:
        ram["usage"] = safe_usage(ram["used"], ram["total"])

    return ram


def process_speedtest(data: dict[str, Any]) -> dict[str, Any]:
    """Process Speedtest data."""

    speedtest: dict[str, Any] = {}

    # Convert the data
    for keys in MAP_SPEEDTEST:
        key, key_to_use, method = safe_unpack_keys(keys)
        state_value = data.get(key)
        if state_value:
            speedtest[key_to_use] = run_method(state_value, method)

    # Get detailed result
    test_result = speedtest.pop("result", None)

    # Get the last speedtest step
    last_run_step: Optional[dict[str, Any]] = process_speedtest_last_step(
        test_result
    )

    # Save the last tested data directly in the speedtest dict
    for key, value in last_run_step.items():
        speedtest[key] = value

    # Convert the timestamp to UTC
    speedtest["timestamp"] = safe_datetime(speedtest.get("timestamp"))

    # Rename servers list
    if "download" in speedtest:
        speedtest["download"]["server_list"] = speedtest["download"].pop(
            "servers", None
        )

    # convert bandwidth from Bps to bps
    for speed in ("download", "upload"):
        if speed in speedtest:
            speedtest[speed]["bandwidth"] = 8 * speedtest[speed].get(
                "bandwidth", 0
            )

    # Remove extra values
    speedtest.pop("type", None)

    return {
        "data": speedtest,
        "result": test_result,
    }


def process_speedtest_last_step(
    data: Optional[list[dict[str, Any]]],
) -> dict[str, Any]:
    """Process Speedtest progress data."""

    last_step: dict[str, Any] = {}

    # Check the length of the data
    if not data or len(data) < 2:
        return last_step

    # Check the last step
    last_step = data[-2]

    # If the last step is a result, then the test is finished
    # and we already have all the data
    if last_step.get("type") != "result":
        return {}

    return last_step


def process_vpnc(
    data: dict[str, Any],
) -> Tuple[dict[AsusVPNType, dict[int, Any]], str]:
    """Process VPNC data."""

    vpnc = {}

    # Get client list
    vpnc_clientlist = (
        data.get("vpnc_clientlist", "")
        .replace("&#62", ">")
        .replace("&#60", "<")
    )
    if vpnc_clientlist != "":
        clients = vpnc_clientlist.split("<")
        vpnc_unit = 0
        for client in clients:
            if client == str():
                continue
            part = client.split(">")
            # Format: name, type, id, login, password, active, vpnc_id, ?, ?, ?, ?, `Web`
            if len(part) < 7:
                continue
            vpnc_id = safe_int(part[6])
            vpnc[vpnc_id] = {
                "type": (
                    AsusVPNType(part[1])
                    if part[1] in [e.value for e in AsusVPNType]
                    else AsusVPNType.UNKNOWN
                ),
                "id": safe_int(part[2]),
                "name": safe_return(part[0]),
                "login": safe_return(part[3]),
                "password": safe_return(part[4]),
                "active": safe_bool(part[5]),
                "vpnc_unit": vpnc_unit,
            }
            vpnc_unit += 1

    # Get clients status
    get_vpnc_status = data.get("get_vpnc_status")
    if get_vpnc_status:
        clients = get_vpnc_status.split("<")
        for client in clients:
            if client == str():
                continue
            part = client.split(">")
            vpnc_id = safe_int(part[2])
            state_code = safe_int(part[0])
            error_code = safe_int(part[1])
            vpnc[vpnc_id].update(
                {
                    "state": (
                        AsusVPNC(state_code)
                        if state_code in [e.value for e in AsusVPNC]
                        else AsusVPNC.UNKNOWN
                    ),
                    "error": (
                        AccessError(error_code)
                        if error_code in [e.value for e in AccessError]
                        else AccessError.UNKNOWN
                    ),
                }
            )

    # Re-sort the data by VPN type / id
    vpn: dict[AsusVPNType, dict[int, Any]] = {
        AsusVPNType.L2TP: {},
        AsusVPNType.OPENVPN: {},
        AsusVPNType.PPTP: {},
        AsusVPNType.SURFSHARK: {},
        AsusVPNType.WIREGUARD: {},
        AsusVPNType.UNKNOWN: {},
    }

    for vpnc_id, info in vpnc.items():
        sorted_id = info.pop("id", None)
        sorted_type = info.pop("type", None)
        info["vpnc_id"] = vpnc_id
        vpn[sorted_type][sorted_id] = info

    # Process WireGuard data
    vpn[AsusVPNType.WIREGUARD] = merge_dicts(
        vpn[AsusVPNType.WIREGUARD], process_vpnc_wireguard(data)
    )
    # Fill missing clients with unknown state
    for num in range(1, 6):
        if num not in vpn[AsusVPNType.WIREGUARD]:
            vpn[AsusVPNType.WIREGUARD][num] = {
                "state": AsusVPNC.UNKNOWN,
                "error": AccessError.NO_ERROR,
            }
        if num not in vpn[AsusVPNType.OPENVPN]:
            vpn[AsusVPNType.OPENVPN][num] = {
                "state": AsusVPNC.UNKNOWN,
                "error": AccessError.NO_ERROR,
            }

    # Remove UNKNOWN VPN type if it's empty
    if not vpn[AsusVPNType.UNKNOWN]:
        vpn.pop(AsusVPNType.UNKNOWN, None)

    return vpn, vpnc_clientlist


def process_vpnc_wireguard(data: dict[str, Any]) -> dict[int, dict[str, Any]]:
    """Process VPNC WireGuard data."""

    wireguard = {}

    for num in range(1, 6):
        client = {}
        for keys in MAP_VPNC_WIREGUARD:
            key, key_to_use, method = safe_unpack_keys(keys)
            state_value = data.get(f"wgc{num}_{key}")
            if state_value:
                client[key_to_use] = run_method(state_value, method)
        if client:
            wireguard[num] = client

    return wireguard


def process_wan(data: dict[str, Any]) -> dict[Any, dict[str, Any]]:
    """Process WAN data."""

    wan: dict[str | int, dict[str, Any]] = {}

    # General data
    wan["internet"] = {}
    for keys in MAP_WAN:
        key, key_to_use, method = safe_unpack_keys(keys)
        state_value = data.get(key)
        if state_value is not None:
            wan["internet"][key_to_use] = run_method(state_value, method)

    # Per-interface data
    for num in (0, 1):
        interface = {}
        for keys in MAP_WAN_ITEM:
            key, key_to_use, method = safe_unpack_keys(keys)
            state_value = data.get(f"wan{num}_{key}")
            if state_value is not None:
                interface[key_to_use] = run_method(state_value, method)
        for extra, extra_key in zip(("", "x"), ("main", "extra")):
            interface[extra_key] = {}
            for keys in MAP_WAN_ITEM_X:
                key, key_to_use, method = safe_unpack_keys(keys)
                state_value = data.get(f"wan{num}_{extra}{key}")
                if state_value is not None:
                    interface[extra_key][key_to_use] = run_method(
                        state_value, method
                    )
        if interface:
            wan[num] = interface

    # Re-sort needed values
    wan[0]["link"] = wan["internet"].pop("link_0", None)
    wan[1]["link"] = wan["internet"].pop("link_1", None)

    # Check state on links
    for num in (0, 1):
        if wan[num].get("link") == ConnectionState.DISCONNECTED:
            wan[num]["state"] = ConnectionStatus.DISCONNECTED

    # WAN aggregation
    aggregation_state = wan["internet"].pop("aggregation_state", None)
    if aggregation_state is not None:
        wan["aggregation"] = {
            "state": aggregation_state,
            "ports": wan["internet"].pop("aggregation_ports", None),
        }

    # Dual WAN
    dualwan_mode = wan["internet"].pop("dualwan_mode", None)
    dualwan_priority = wan["internet"].pop("dualwan_priority", None)
    if dualwan_mode is not None and dualwan_priority is not None:
        wan["dualwan"] = {
            "mode": dualwan_mode,
            "priority": dualwan_priority,
        }
        # Check whether it is actually active
        if wan["dualwan"]["priority"][1] == "none":
            wan["dualwan"]["priority"][1] = None
            wan["dualwan"]["state"] = False
        else:
            wan["dualwan"]["state"] = True

    # Assign main IP address
    wan["internet"]["ip_address"] = wan[wan["internet"]["unit"]]["main"][
        "ip_address"
    ]

    return wan


def process_wireguard_server(
    data: dict[str, Any],
) -> dict[int, dict[str, Any]]:
    """Process WireGuard data."""

    wireguard = {}

    # Server data
    for keys in MAP_WIREGUARD_SERVER:
        key, key_to_use, method = safe_unpack_keys(keys)
        state_value = data.get(key)
        if state_value:
            wireguard[key_to_use] = run_method(state_value, method)

    # Per-client data
    wireguard["clients"] = {}
    for num in range(1, 11):
        client = {}
        for keys in MAP_WIREGUARD_CLIENT:
            key, key_to_use, method = safe_unpack_keys(keys)
            state_value = data.get(f"wgs1_c{num}_{key}")
            if state_value:
                client[key_to_use] = run_method(state_value, method)
        if client:
            wireguard["clients"][num] = client

    if "status" in wireguard:
        status = wireguard["status"].get("client_status")
        if status:
            for client in status:
                if client.get("index") in wireguard["clients"]:
                    wireguard["clients"][client.get("index")]["state"] = (
                        ConnectionState(client.get("status"))
                    )

    # Remove the `status` value
    wireguard.pop("status", None)

    return {1: wireguard}


def process_wlan(
    data: dict[str, Any], wlan_list: list[Wlan]
) -> dict[str, Any]:
    """Process WLAN data."""

    wlan = {}

    for interface in wlan_list:
        index = list(Wlan).index(interface)
        info = {}
        for pair in MAP_WLAN:
            key, method = safe_unpack_key(pair)
            info[key.format(index)[4:]] = (
                method(data.get(key.format(index)))
                if method
                else data.get(key.format(index))
            )
        wlan[interface.value] = info

    return wlan

def process_dsl(dsl_info: dict[str, Any]) -> dict[str, Any]:
    """Process DSL data"""

    def remove_units(value: Optional[str]) -> Optional[str]:
        """Remove units from the value."""

        if value is None:
            return None
        return value.split(" ")[0]

    dsl: dict[str, Any] = {}

    # Data preprocessing
    _dataratedown = remove_units(dsl_info.get("dsllog_dataratedown"))
    _datarateup = remove_units(dsl_info.get("dsllog_datarateup"))

    dsl["datarate"] = {
        "down": safe_int(_dataratedown, 0),
        "up": safe_int(_datarateup, 0),
    }

    return dsl



"""Firmware endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.const import UNKNOWN_MEMBER
from asusrouter.modules.data import AsusData
from asusrouter.modules.firmware import (
    Firmware,
    WebsError,
    WebsFlag,
    WebsUpdate,
    WebsUpgrade,
)
from asusrouter.tools.converters import clean_string, safe_enum, safe_int
from asusrouter.tools.readers import read_js_variables as read  # noqa: F401

REQUIRE_FIRMWARE = True


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process firmware data."""

    # Stay on the safe side with Enums
    # since the endpoint can provide empty strings
    _available = Firmware(data.get("webs_state_info", "")).safe()
    _available_beta = Firmware(data.get("webs_state_info_beta", "")).safe()

    # Load all the static data
    firmware: dict[str, Any] = {
        "current": None,
        "state": False,
        "available": None,
        "state_beta": False,
        "available_beta": None,
        "webs": {
            "update": safe_enum(
                WebsUpdate,
                safe_int(data.get("webs_state_update")),
                default=WebsUpdate.UNKNOWN,
            ),
            "upgrade": safe_enum(
                WebsUpgrade,
                safe_int(data.get("webs_state_upgrade")),
                default=WebsUpgrade.INACTIVE,
            ),
            "available": _available,
            "available_beta": _available_beta,
            "required": Firmware(data.get("webs_state_REQinfo", "")).safe(),
            "error": safe_enum(
                WebsError,
                safe_int(data.get("webs_state_error")),
                default_value=UNKNOWN_MEMBER,
            ),
            "flag": safe_enum(
                WebsFlag,
                safe_int(data.get("webs_state_flag")),
                default_value=UNKNOWN_MEMBER,
            ),
            "level": safe_int(data.get("webs_state_level")),
        },
        "cfg": {
            "check": safe_int(data.get("cfg_check")),
            "upgrade": safe_int(data.get("cfg_upgrade")),
        },
        "sig": {
            "update": safe_int(data.get("sig_state_update")),
            "upgrade": safe_int(data.get("sig_state_upgrade")),
            "version": clean_string(data.get("sig_ver")),
            "error": safe_int(data.get("sig_state_error")),
            "flag": safe_int(data.get("sig_state_flag")),
        },
        "hndwr": {
            "status": safe_int(data.get("hndwr_status")),
        },
    }

    # Check the current firmware
    _current = data.get("firmware")
    firmware["current"] = _current

    # Check if the stable firmware is available
    firmware["state"] = (
        _current < _available if _current else _available is not None
    )
    if firmware["state"]:
        firmware["available"] = _available
    # Check if the beta firmware is available
    firmware["state_beta"] = (
        _current < _available_beta if _current else _available_beta is not None
    )
    if firmware["state_beta"]:
        firmware["available_beta"] = _available_beta

    return {AsusData.FIRMWARE: firmware}



"""IP address module."""

import re
from enum import Enum
from typing import Optional

from asusrouter.tools.converters import clean_string


class IPAddressType(str, Enum):
    """IP address type class."""

    DHCP = "dhcp"
    PPPOE = "pppoe"
    PPTP = "pptp"
    STATIC = "static"
    UNKNOWN = "unknown"


def read_ip_address_type(data: Optional[str]) -> IPAddressType:
    """Read IP address type from data string."""

    data = clean_string(data)

    if data:
        data = data.lower()

    # Find the IP address type in the IPAddressType enum by value
    for ip_address_type in IPAddressType:
        if ip_address_type.value == data:
            return ip_address_type

    # Some other values to convert
    if data == "manual":
        return IPAddressType.STATIC

    # If the IP address type is not found, return the UNKNOWN type
    return IPAddressType.UNKNOWN


def read_dns_ip_address(data: str) -> list[str]:
    """Read DNS record and return the list of IP addresses."""

    # Regex to find each IP address in a string of IP addresses
    regex = r"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"
    # Find all IP addresses in the string
    matches = re.findall(regex, data)

    return matches



"""LED module."""

from __future__ import annotations

from enum import IntEnum
from typing import Any, Awaitable, Callable

from asusrouter.modules.endpoint import Endpoint


class AsusLED(IntEnum):
    """Asus LED state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusLED,
    **kwargs: Any,
) -> bool:
    """Set the LED state."""

    # Prepare the arguments
    arguments = {"led_val": state.value}

    # Run the service
    return await callback(
        service="start_ctrl_led",
        arguments=arguments,
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )


async def keep_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusLED = AsusLED.ON,
    **kwargs: Any,
) -> bool:
    """Keep the LED state."""

    identity = kwargs.get("identity")

    # Check if identity is available and if endpoints are defined
    if identity is None or not identity.endpoints:
        return False

    # Get the sysinfo
    sysinfo = identity.endpoints.get(Endpoint.SYSINFO)

    # Only when LEDs are off and if sysinfo is available
    if not sysinfo or state == AsusLED.ON:
        return False

    # Toggle the LED
    await set_state(callback, AsusLED.ON, **kwargs)
    await set_state(callback, AsusLED.OFF, **kwargs)

    return True



"""Client module for AsusRouter."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import Any, Optional

from asusrouter.modules.connection import (
    ConnectionState,
    ConnectionType,
    InternetMode,
    get_connection_type,
    get_internet_mode,
)
from asusrouter.modules.const import MapValueType
from asusrouter.modules.ip_address import IPAddressType, read_ip_address_type
from asusrouter.modules.vendor import replace_vendor
from asusrouter.tools.converters import (
    safe_bool,
    safe_float,
    safe_int,
    safe_time_from_delta,
    safe_unpack_key,
)


@dataclass
class AsusClient:
    """Client class."""

    state: ConnectionState = ConnectionState.UNKNOWN

    description: Optional[AsusClientDescription] = None

    connection: Optional[AsusClientConnection] = None


@dataclass
class AsusClientConnection:
    """Connection class."""

    type: ConnectionType = ConnectionType.DISCONNECTED
    ip_address: Optional[str] = None
    ip_method: Optional[str] = None
    internet_state: Optional[bool] = None
    internet_mode: InternetMode = InternetMode.UNKNOWN
    node: Optional[str] = None
    online: bool = False

    # Is an AiMesh node
    aimesh: Optional[bool] = None


@dataclass
class AsusClientConnectionWlan(AsusClientConnection):
    """WLAN connection class."""

    guest: Optional[bool] = None
    guest_id: Optional[int] = None
    rssi: Optional[int] = None
    since: Optional[datetime] = None
    rx_speed: Optional[float] = None
    tx_speed: Optional[float] = None


@dataclass
class AsusClientDescription:
    """Client description class."""

    name: Optional[str] = None
    mac: Optional[str] = None

    vendor: Optional[str] = None


CLIENT_MAP: dict[str, list[MapValueType]] = {
    "connected_since": [
        ("wlConnectTime", safe_time_from_delta),
    ],
    "connection_type": [
        ("connection_type"),
        ("isWL", safe_int),
    ],
    "guest": [
        ("guest"),
        ("isGN", safe_int),
    ],
    "internet_mode": [
        ("internetMode", get_internet_mode),
    ],
    "internet_state": [
        ("internetState", safe_bool),
    ],
    "ip": [
        ("ip"),
    ],
    "ip_method": [
        ("ipMethod"),
    ],
    "mac": [
        ("mac"),
    ],
    "name": [
        ("nickName"),
        ("name"),
        ("mac"),
    ],
    "node": [
        ("node"),
    ],
    "online": [
        ("online"),
        ("isOnline", safe_bool),
    ],
    "rssi": [
        ("rssi", safe_int),
    ],
    "rx_speed": [
        ("curRx", safe_float),
    ],
    "tx_speed": [
        ("curTx", safe_float),
    ],
}

CLIENT_MAP_DESCRIPTION: dict[str, list[MapValueType]] = {
    "name": [
        ("nickName"),
        ("name"),
        ("mac"),
    ],
    "mac": [
        ("mac"),
    ],
    "vendor": [
        ("vendor", replace_vendor),
    ],
}

CLIENT_MAP_CONNECTION: dict[str, list[MapValueType]] = {
    "type": [
        ("connection_type", get_connection_type),
        ("isWL", get_connection_type),
    ],
    "ip_address": [
        ("ip"),
    ],
    "ip_method": [
        ("ipMethod", read_ip_address_type),
    ],
    "internet_state": [
        ("internetState", safe_bool),
    ],
    "internet_mode": [
        ("internetMode", get_internet_mode),
    ],
    "node": [
        ("node"),
    ],
    "online": [
        ("online", safe_bool),
        ("isOnline", safe_bool),
    ],
    "aimesh": [
        ("amesh_isRe", safe_bool),
    ],
}

CLIENT_MAP_CONNECTION_WLAN: dict[str, list[MapValueType]] = {
    "guest_id": [
        ("guest"),
        ("isGN", safe_int),
    ],
    "rssi": [
        ("rssi", safe_int),
    ],
    "since": [
        ("wlConnectTime", safe_time_from_delta),
    ],
    "rx_speed": [
        ("curRx", safe_float),
    ],
    "tx_speed": [
        ("curTx", safe_float),
    ],
}


def process_client(
    data: dict[str, Any], history: Optional[AsusClient] = None, **kwargs: Any
) -> AsusClient:
    """
    Process client data.

    Parameters:
    data (dict[str, Any]): The client data to process.
    history (Optional[AsusClient]): The previous state of the client, if any.

    Returns:
    AsusClient: The processed client data.
    """

    description: AsusClientDescription = process_client_description(data)
    connection: AsusClientConnection | AsusClientConnectionWlan = (
        process_client_connection(data)
    )

    state: ConnectionState = process_client_state(connection, **kwargs)

    # Clean disconnected client
    if state != ConnectionState.CONNECTED:
        connection = process_disconnected(connection)

    # Check history for the oscillating values
    if history is not None and history.connection is not None:
        connection = process_history(connection, history.connection)

    return AsusClient(state=state, description=description, connection=connection)


def process_data(data: dict[str, Any], mapping: dict[str, Any], obj: Any) -> Any:
    """Process data based on a mapping and set attributes on an object."""

    # Go through all keys in mapping
    for key, value in mapping.items():
        for pair in value:
            # Get the search key and converter(s)
            key_to_find, converters = safe_unpack_key(pair)
            # Get the value from the data
            item = data.get(key_to_find)

            # Process the value if it's an actual value
            if item is not None and item != str():
                # Apply converters one by one if there are multiple
                if isinstance(converters, list):
                    for converter in converters:
                        item = converter(item)
                # Apply single converter
                elif converters:
                    item = converters(item)

                # Set the attribute on the object
                setattr(obj, key, item)

                # We found the value, no need to continue searching
                break

    return obj


def process_client_description(data: dict[str, Any]) -> AsusClientDescription:
    """Process client description data."""

    return process_data(data, CLIENT_MAP_DESCRIPTION, AsusClientDescription())


def process_client_connection(data: dict[str, Any]) -> AsusClientConnection:
    """Process client connection data."""

    connection = process_data(data, CLIENT_MAP_CONNECTION, AsusClientConnection())

    if not connection.type in (ConnectionType.WIRED, ConnectionType.DISCONNECTED):
        connection = process_client_connection_wlan(data, connection)

    return connection


def process_client_connection_wlan(
    data: dict[str, Any], base_connection: AsusClientConnection
) -> AsusClientConnectionWlan:
    """Process WLAN client connection data."""

    wlan_connection = AsusClientConnectionWlan(**base_connection.__dict__)
    wlan_connection = process_data(data, CLIENT_MAP_CONNECTION_WLAN, wlan_connection)

    # Mark `guest` attribute if `guest_id` is non-zero
    wlan_connection.guest = (
        wlan_connection.guest_id != 0 and wlan_connection.guest_id is not None
    )

    return wlan_connection


def process_client_state(
    connection: AsusClientConnection,
    **kwargs: Any,
) -> ConnectionState:
    """Process client state."""

    # If no IP address or type is disconnected, it's disconnected
    if connection.ip_address is None or connection.type == ConnectionType.DISCONNECTED:
        return ConnectionState.DISCONNECTED

    # If client is an AiMesh node, it's not a client
    if connection.aimesh is True:
        return ConnectionState.DISCONNECTED

    # If we have support for AiMesh but no node, client is disconnected
    aimesh = kwargs.get("aimesh", False)
    if aimesh is True and connection.node is None:
        return ConnectionState.DISCONNECTED

    # This one is a weak check, should always be the last one
    if connection.online is True:
        return ConnectionState.CONNECTED

    return ConnectionState.DISCONNECTED


def process_disconnected(connection: AsusClientConnection) -> AsusClientConnection:
    """Process disconnected client."""

    # We keep values set by the user:
    # - internet mode
    # - ip method
    # - ip address if ip method is static
    return AsusClientConnection(
        type=ConnectionType.DISCONNECTED,
        ip_address=connection.ip_address
        if connection.ip_method == IPAddressType.STATIC
        else None,
        ip_method=connection.ip_method,
        internet_mode=connection.internet_mode,
    )


def process_history(
    connection: AsusClientConnection, history: AsusClientConnection
) -> AsusClientConnection:
    """Process values from history to avoid oscillating values."""

    # Wireless connection
    if isinstance(connection, AsusClientConnectionWlan) and isinstance(
        history, AsusClientConnectionWlan
    ):
        # Fix the oscillating connection time ignore any value less than 10 seconds
        # This comes from the device itself and cannot be fixed from AsusRouter side
        if connection.since is not None and history.since is not None:
            diff = abs((history.since - connection.since).total_seconds())
            if diff < 10:
                connection.since = history.since

    return connection



"""Data module."""

from __future__ import annotations

import asyncio
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Any, Optional

from asusrouter.tools.converters import safe_bool


class AsusData(str, Enum):
    """AsusRouter data class."""

    AIMESH = "aimesh"
    AURA = "aura"
    BOOTTIME = "boottime"
    CLIENTS = "clients"
    CPU = "cpu"
    DEVICEMAP = "devicemap"
    DSL = "dsl"
    FIRMWARE = "firmware"
    FIRMWARE_NOTE = "firmware_note"
    FLAGS = "flags"
    GWLAN = "gwlan"
    LED = "led"
    NETWORK = "network"
    NODE_INFO = "node_info"
    OPENVPN = "openvpn"
    OPENVPN_CLIENT = "openvpn_client"
    OPENVPN_SERVER = "openvpn_server"
    PARENTAL_CONTROL = "parental_control"
    PING = "ping"
    PORT_FORWARDING = "port_forwarding"
    PORTS = "ports"
    RAM = "ram"
    SPEEDTEST = "speedtest"
    # SPEEDTEST_HISTORY = "speedtest_history"
    SPEEDTEST_RESULT = "speedtest_result"
    # SPEEDTEST_SERVERS = "speedtest_servers"
    SYSINFO = "sysinfo"
    SYSTEM = "system"
    TEMPERATURE = "temperature"
    VPNC = "vpnc"
    VPNC_CLIENTLIST = "vpnc_clientlist"
    WAN = "wan"
    WIREGUARD = "wireguard"
    WIREGUARD_CLIENT = "wireguard_client"
    WIREGUARD_SERVER = "wireguard_server"
    WLAN = "wlan"


@dataclass
class AsusDataState:
    """State of data."""

    data: Optional[Any] = None
    timestamp: datetime = datetime.now(timezone.utc)
    active: bool = False
    inactive_event: asyncio.Event = asyncio.Event()

    def start(self) -> None:
        """Set to active."""

        self.active = True
        self.inactive_event.clear()

    def stop(self) -> None:
        """Set to not-active."""

        self.active = False
        self.inactive_event.set()

    def update(self, data: Any) -> None:
        """Update the state."""

        self.data = data
        # Set timestamp to the current utc time
        self.timestamp = datetime.now(timezone.utc)
        # Set to inactive
        self.stop()

    def update_state(self, state: Any, last_id: Optional[int] = None) -> None:
        """Update a state variable in the data dict."""

        # Convert the state if needed
        state = convert_state(state)

        if last_id is not None:
            if not isinstance(self.data, dict):
                self.data = {}
            self.data.setdefault(last_id, {})["state"] = state
            return

        if isinstance(self.data, dict):
            self.data["state"] = state
            return

        self.data = {"state": state}

    def offset_time(self, offset: Optional[int]) -> None:
        """Offset the timestamp."""

        if offset is None:
            self.timestamp = datetime.now(timezone.utc)
            return

        self.timestamp = datetime.now(timezone.utc) + timedelta(seconds=offset)


def convert_state(state: Any):
    """Convert the state to a correct one."""

    # If the state is not boolean, convert it to boolean
    if not isinstance(state, bool):
        state = safe_bool(state)

    return state



"""Connection module."""

from enum import Enum, IntEnum
from typing import Optional

from asusrouter.modules.wlan import Wlan
from asusrouter.tools.converters import get_enum_key_by_value, safe_int


class ConnectionState(IntEnum):
    """Connection state enum."""

    UNKNOWN = -999

    DISCONNECTED = 0
    CONNECTED = 1


class ConnectionStatus(IntEnum):
    """Connection status enum."""

    UNKNOWN = -999

    ERROR = -1
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2


class ConnectionType(str, Enum):
    """Connection type class."""

    WIRED = "wired"
    WLAN_2G = Wlan.FREQ_2G.value
    WLAN_5G = Wlan.FREQ_5G.value
    WLAN_5G2 = Wlan.FREQ_5G2.value
    WLAN_6G = Wlan.FREQ_6G.value
    DISCONNECTED = "disconnected"


CONNECTION_TYPE = {
    0: ConnectionType.WIRED,
    1: ConnectionType.WLAN_2G,
    2: ConnectionType.WLAN_5G,
    3: ConnectionType.WLAN_5G2,
    4: ConnectionType.WLAN_6G,
}


class InternetMode(str, Enum):
    """Internet mode class."""

    ALLOW = "allow"
    BLOCK = "block"
    TIME = "time"
    UNKNOWN = "unknown"


def get_internet_mode(value: str) -> InternetMode:
    """Get internet mode."""

    return get_enum_key_by_value(InternetMode, value, InternetMode.UNKNOWN)


def get_connection_type(value: Optional[int]) -> ConnectionType:
    """Get connection type."""

    # Check that it's actually an int
    # This would actually work with float as well doing direct conversion
    value = safe_int(value)

    # ------------------------------------------------------ #
    # I am officially considering this a feature, not a bug. #
    # ------------------------------------------------------ #

    if value is None:
        value = 0

    return CONNECTION_TYPE.get(value, ConnectionType.WIRED)



"""Error endpoint module.

This is not an actual endpoint, but rather a module that is used
to handle errors from any endpoint."""

from __future__ import annotations

import logging
from enum import IntEnum
from typing import Any

from asusrouter.error import AsusRouterAccessError, AsusRouterLogoutError
from asusrouter.modules.endpoint import EndpointType
from asusrouter.tools.readers import read_json_content

_LOGGER = logging.getLogger(__name__)


class AccessError(IntEnum):
    """Access error enum."""

    UNKNOWN = -999

    NO_ERROR = 0
    AUTHORIZATION = 2
    CREDENTIALS = 3
    TRY_AGAIN = 7
    LOGOUT = 8
    ANOTHER = 9
    CAPTCHA = 10
    RESET_REQUIRED = 11


def handle_access_error(
    endpoint: EndpointType, status: Any, headers: Any, content: Any
) -> None:
    """Handle access errors."""

    # Read the page as json
    message = read_json_content(content)
    _LOGGER.debug("Access error message: %s", message)

    # Get error code
    error_status = int(message.get("error_status") or -999)
    try:
        error = AccessError(error_status)
    except ValueError:
        error = AccessError.UNKNOWN

    # Additional values
    attributes: dict[str, Any] = {}

    # Handle logout code (even though it's not an error)
    if error == AccessError.LOGOUT:
        raise AsusRouterLogoutError("Session is logged out")

    # Try again later error
    if error == AccessError.TRY_AGAIN:
        timeout = message.get("remaining_lock_time")
        if timeout:
            attributes["timeout"] = int(timeout)

    # Return
    raise AsusRouterAccessError(
        "Access error",
        error,
        attributes,
    )



"""Endpoint module for AsusRouter."""

from __future__ import annotations

import importlib
import logging
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from types import ModuleType
from typing import Any, Awaitable, Callable, Optional, Union

from asusrouter.const import RequestType
from asusrouter.error import AsusRouter404Error
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.firmware import Firmware
from asusrouter.modules.wlan import Wlan

_LOGGER = logging.getLogger(__name__)


class Endpoint(str, Enum):
    """Endpoint enum. These endpoints are used to receive data from the device."""

    DEVICEMAP = "ajax_status.xml"
    DSL = "ajax_AdslStatus.asp"
    ETHERNET_PORTS = "ajax_ethernet_ports.asp"
    FIRMWARE = "detect_firmware.asp"
    FIRMWARE_NOTE = "release_note0.asp"
    FIRMWARE_NOTE_AIMESH = "release_note_amas.asp"
    HOOK = "appGet.cgi"
    NETWORKMAPD = "update_networkmapd.asp"
    ONBOARDING = "ajax_onboarding.asp"
    PORT_STATUS = "get_port_status.cgi"
    STATE = "state.js"
    SYSINFO = "ajax_sysinfo.asp"
    TEMPERATURE = "ajax_coretmp.asp"
    UPDATE_CLIENTS = "update_clients.asp"
    VPN = "ajax_vpn_status.asp"
    # Endpoints known but not used / not available on test devices
    # RGB = "light_effect.html"


class EndpointControl(str, Enum):
    """Control endpoint enum. These endpoints are used to set parameters to the device."""

    APPLY = "apply.cgi"
    COMMAND = "applyapp.cgi"


class EndpointService(str, Enum):
    """Service endpoints."""

    LOGIN = "login.cgi"
    LOGOUT = "Logout.asp"


class EndpointTools(str, Enum):
    """Tools endpoints."""

    # AURA control / RGB
    AURA = "set_ledg.cgi"
    # Network tools: ping, jitter, loss
    NETWORK = "netool.cgi"


class EndpointNoCheck(str, Enum):
    """Endpoints that should not be checked for availability."""

    # AURA control / RGB
    AURA = "set_ledg.cgi"


# Typehint for the endpoint
EndpointType = Union[Endpoint, EndpointControl, EndpointService, EndpointTools]

# Force request type for the endpoint
ENDPOINT_FORCE_REQUEST = {
    Endpoint.PORT_STATUS: RequestType.GET,
    EndpointTools.NETWORK: RequestType.GET,
}


def _get_module(
    endpoint: Endpoint | EndpointControl | EndpointTools,
) -> Optional[ModuleType]:
    """Attempt to get the module for the endpoint."""

    try:
        # Get the module name from the endpoint
        module_name = f"asusrouter.modules.endpoint.{endpoint.name.lower()}"

        # Import the module in a separate thread to avoid blocking the main thread
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(importlib.import_module, module_name)
            submodule = future.result()

        # Return the module
        return submodule
    except ModuleNotFoundError:
        _LOGGER.debug("No module found for endpoint %s", endpoint)
        return None


def read(
    endpoint: Endpoint | EndpointControl | EndpointTools, content: str
) -> dict[str, Any]:
    """Read the data from an endpoint."""

    _LOGGER.debug("Reading data from endpoint %s", endpoint)

    # Get the module
    submodule = _get_module(endpoint)

    # Read the data if module found
    if submodule:
        return submodule.read(content)

    return {}


def process(
    endpoint: Endpoint,
    data: dict[str, Any],
    history: Optional[dict[AsusData, AsusDataState]] = None,
    firmware: Optional[Firmware] = None,
    wlan: Optional[list[Wlan]] = None,
) -> dict[AsusData, Any]:
    """Process the data from an endpoint."""

    _LOGGER.debug("Processing data from endpoint %s", endpoint)

    # Get the module
    submodule = _get_module(endpoint)

    # Process the data if module found
    if submodule:
        # Check if the submodule requires history
        require_history = getattr(submodule, "REQUIRE_HISTORY", False)
        if require_history:
            data_set(data, history=history)
        # Check if the submodule requires identity
        require_firmware = getattr(submodule, "REQUIRE_FIRMWARE", False)
        if require_firmware:
            data_set(data, firmware=firmware)
        # Check if the submodule requires wlan
        require_wlan = getattr(submodule, "REQUIRE_WLAN", False)
        if require_wlan:
            data_set(data, wlan=wlan)

        # Process the data
        return submodule.process(data)

    return {}


def data_set(data: dict[str, Any], **kwargs: Any) -> dict[str, Any]:
    """Append the data to the data dict."""

    # Go through the kwargs and append the data
    for key, value in kwargs.items():
        data[key] = value

    # Return the data
    return data


def data_get(data: dict[str, Any], key: str) -> Optional[Any]:
    """Extract value from the data dict and update the data dict."""

    # Get the value
    value = data.get(key, None)

    # Remove the value from the data dict
    data.pop(key, None)

    # Return the value
    return value


async def check_available(
    endpoint: Endpoint | EndpointTools,
    api_query: Callable[..., Awaitable[Any]],
) -> tuple[bool, Optional[Any]]:
    """Check whether the endpoint is available or returns 404."""

    try:
        status, _, content = await api_query(endpoint)
        if status == 200:
            return (True, content)
    except AsusRouter404Error:
        return (False, None)

    return (False, None)



"""State module."""

from __future__ import annotations

import importlib
import logging
from enum import Enum
from types import ModuleType
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.aura import AsusAura
from asusrouter.modules.connection import ConnectionState
from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.parental_control import (
    AsusBlockAll,
    AsusParentalControl,
    ParentalControlRule,
)
from asusrouter.modules.port_forwarding import AsusPortForwarding
from asusrouter.modules.system import AsusSystem
from asusrouter.modules.vpnc import AsusVPNC
from asusrouter.modules.wireguard import (
    AsusWireGuardClient,
    AsusWireGuardServer,
)
from asusrouter.modules.wlan import AsusWLAN
from asusrouter.tools.converters import get_enum_key_by_value

from .led import AsusLED
from .openvpn import AsusOVPNClient, AsusOVPNServer

_LOGGER = logging.getLogger(__name__)


class AsusStateNone(int, Enum):
    """Asus state none."""

    NONE = 0


# AsusState = Union[AsusLED, AsusOVPNClient, AsusOVPNServer, AsusStateNone]


class AsusState(Enum):
    """Asus state."""

    NONE = AsusStateNone
    AURA = AsusAura
    BLOCK_ALL = AsusBlockAll
    CONNECTION = ConnectionState
    LED = AsusLED
    OPENVPN_CLIENT = AsusOVPNClient
    OPENVPN_SERVER = AsusOVPNServer
    PARENTAL_CONTROL = AsusParentalControl
    PC_RULE = ParentalControlRule
    PORT_FORWARDING = AsusPortForwarding
    SYSTEM = AsusSystem
    VPNC = AsusVPNC
    WIREGUARD_CLIENT = AsusWireGuardClient
    WIREGUARD_SERVER = AsusWireGuardServer
    WLAN = AsusWLAN


AsusStateMap: dict[AsusState, Optional[AsusData]] = {
    AsusState.NONE: None,
    AsusState.AURA: AsusData.AURA,
    AsusState.BLOCK_ALL: AsusData.PARENTAL_CONTROL,
    AsusState.CONNECTION: None,
    AsusState.LED: AsusData.LED,
    AsusState.OPENVPN_CLIENT: AsusData.OPENVPN_CLIENT,
    AsusState.OPENVPN_SERVER: AsusData.OPENVPN_SERVER,
    AsusState.PARENTAL_CONTROL: AsusData.PARENTAL_CONTROL,
    AsusState.PC_RULE: AsusData.PARENTAL_CONTROL,
    AsusState.PORT_FORWARDING: AsusData.PORT_FORWARDING,
    AsusState.SYSTEM: AsusData.SYSTEM,
    AsusState.VPNC: AsusData.VPNC,
    AsusState.WIREGUARD_CLIENT: AsusData.WIREGUARD_CLIENT,
    AsusState.WIREGUARD_SERVER: AsusData.WIREGUARD_SERVER,
    AsusState.WLAN: AsusData.WLAN,
}


def add_conditional_state(state: AsusState, data: AsusData) -> None:
    """A callback to add / change AsusStateMap."""

    if not isinstance(state, AsusState) or not isinstance(data, AsusData):
        _LOGGER.debug("Invalid state or data type: %s -> %s", state, data)
        return

    AsusStateMap[state] = data
    _LOGGER.debug("Added conditional state rule: %s -> %s", state, data)


def get_datatype(state: Optional[Any]) -> Optional[AsusData]:
    """Get the datatype."""

    asus_state = get_enum_key_by_value(
        AsusState, type(state), default=AsusState.NONE
    )

    return AsusStateMap.get(asus_state)


def _get_module_name(state: AsusState) -> Optional[str]:
    """Get the module name."""

    module_class = get_datatype(state)
    if module_class:
        return module_class.value

    return None


def _get_module(state: AsusState) -> Optional[ModuleType]:
    """Get the module."""

    # Module name
    module_name = _get_module_name(state)
    if not module_name:
        return None

    if module_name.endswith(("_client", "_server")):
        module_name = module_name[:-7]

    # Module path
    module_path = f"asusrouter.modules.{module_name}"

    try:
        # Import the module
        submodule = importlib.import_module(module_path)
        # Return the module
        return submodule
    except ModuleNotFoundError:
        _LOGGER.debug("No module found for state %s", state)
        return None


def _has_method(module: ModuleType, method: str) -> bool:
    """Check if the module has the method."""

    return hasattr(module, method) and callable(getattr(module, method))


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusState,
    **kwargs: Any,
) -> bool:
    """Set the state."""

    # Get the module
    submodule = _get_module(state)

    # Process the data if module found
    if submodule and _has_method(submodule, "set_state"):
        # Determine the extra parameter
        if getattr(submodule, "REQUIRE_STATE", False):
            kwargs["extra_param"] = kwargs.get("router_state")
        if getattr(submodule, "REQUIRE_IDENTITY", False):
            kwargs["extra_param"] = kwargs.get("identity")

        # Call the function with the determined parameters
        return await submodule.set_state(
            callback=callback,
            state=state,
            **kwargs,
        )

    return False


def save_state(
    state: AsusState,
    library: dict[AsusData, AsusDataState],
    needed_time: Optional[int] = None,
    last_id: Optional[int] = None,
) -> None:
    """Save the state."""

    # Get the correct data key
    datatype = get_datatype(state)
    if datatype is None or datatype not in library:
        return

    # Save the state
    library[datatype].update_state(state, last_id)
    library[datatype].offset_time(needed_time)


async def keep_state(
    callback: Callable[..., Awaitable[Any]],
    states: Optional[AsusState | list[AsusState]],
    **kwargs: Any,
) -> None:
    """Keep the state."""

    if states is None:
        return

    # Make sure the state is a list
    states = [states] if not isinstance(states, list) else states

    # Process each state
    awaitables = [
        submodule.keep_state(callback, state, **kwargs)
        for state in states
        if (submodule := _get_module(state))
        and _has_method(submodule, "keep_state")
    ]

    # Execute all awaitables
    for awaitable in awaitables:
        await awaitable



"""Legacy tools module."""

from __future__ import annotations

from typing import Optional

from asusrouter.modules.port_forwarding import (
    KEY_PORT_FORWARDING_LIST,
    PortForwardingRule,
)


def compile_port_forwarding(data: list[PortForwardingRule]) -> Optional[dict[str, str]]:
    """Compile port forwarding rules"""

    if not isinstance(data, list):
        return None

    result = ""
    for rule in data:
        name = str()
        if rule.name is not None and rule.name != "None":
            name = rule.name
        port_external = str()
        if rule.port_external is not None and rule.port_external != "None":
            port_external = rule.port_external
        ip = str()
        if rule.ip_address is not None and rule.ip_address != "None":
            ip = rule.ip_address
        port = str()
        if rule.port is not None and rule.port != "None":
            port = rule.port
        protocol = str()
        if rule.protocol is not None and rule.protocol != "None":
            protocol = rule.protocol
        ip_external = str()
        if rule.ip_external is not None and rule.ip_external != "None":
            ip_external = rule.ip_external

        result += f"<{name}>{port_external}>{ip}>{port}>{protocol}>{ip_external}>"

    return {
        KEY_PORT_FORWARDING_LIST: result,
    }



"""Constants module for AsusRouter"""

from enum import Enum


# Enums
class ContentType(str, Enum):
    """Content type enum."""

    UNKNOWN = "unknown"

    BINARY = "application/octet-stream"
    HTML = "text/html"
    JSON = "application/json"
    TEXT = "text/plain"
    XML = "application/xml"


class RequestType(str, Enum):
    """Request type enum."""

    GET = "get"
    POST = "post"


# Asus constants
USER_AGENT = "asusrouter--DUTUtil-"

# Library defaults
DEFAULT_CACHE_TIME = 5.0
DEFAULT_SLEEP_TIME = 0.1
DEFAULT_TIMEOUT = 15
DEFAULT_TIMEOUT_CONNECTION = 180.0

DEFAULT_RESULT_SUCCESS = {"statusCode": "200"}

UNKNOWN_MEMBER = -999

# --------------------
# Value maps -->
# --------------------
# These maps are used to read data from the device
# into the correct variables and apply a converter if needed


# --------------------
# <-- Value maps
# --------------------



"""Firmware module."""

from __future__ import annotations

import logging
import re
from enum import Enum
from typing import Any, Optional

from asusrouter.tools.converters import clean_string, safe_int

_LOGGER = logging.getLogger(__name__)


class FirmwareType(Enum):
    """Type of firmware."""

    UNKNOWN = -999
    STOCK = 0
    MERLIN = 1
    GNUTON = 2


class WebsError(Enum):
    """Webs Upgrade error status."""

    UNKNOWN = -999
    NONE = 0
    DOWNLOAD_ERROR = 1
    SPACE_ERROR = 2
    FW_ERROR = 3


class WebsFlag(Enum):
    """Webs flag status."""

    UNKNOWN = -999
    DONT = 0  # Don't do upgrade
    AVAILABLE = 1  # New FW available
    FORCE = 2  # Force upgrade


class WebsUpdate(Enum):
    """Webs Update status."""

    UNKNOWN = -999
    ACTIVE = 0
    INACTIVE = 1


class WebsUpgrade(Enum):
    """Webs Upgrade status."""

    UNKNOWN = -999
    INACTIVE = -1
    DOWNLOADING = 0
    FINISHED = 1
    ACTIVE = 2


class Firmware:
    """Firmware class.

    This class contains information about the firmware of a device."""

    _fw_warned: list[str] = []

    def __init__(
        self,
        version: Optional[str] = None,
        major: Optional[str] = None,
        minor: Optional[int] = None,
        build: Optional[int] = None,
        revision: Optional[int | str] = None,
    ) -> None:
        """Initialize the firmware object."""

        self.major: Optional[str] = None
        self.minor: Optional[int] = None
        self.build: Optional[int] = None
        self.revision: Optional[int | str] = None

        self.source: FirmwareType = FirmwareType.STOCK
        self.rog: bool = False
        self.beta: bool = False

        if version:
            self.from_string(version)
        else:
            self.major = major
            self.minor = minor
            self.build = build
            self.revision = revision
            self._update_rog()

        self._update_beta()
        self._update_source()

    def safe(self) -> Optional[Firmware]:
        """Return self if exists, otherwise None."""

        if self.major:
            return self
        return None

    def _update_beta(self) -> None:
        """Update the beta flag."""

        self.beta = False
        if self.major:
            self.beta = self.major[0] == "9"

    def _update_rog(self) -> None:
        """Update the ROG flag."""

        self.rog = False
        if isinstance(self.revision, str):
            self.rog = "_rog" in self.revision
            # Remove the _rog suffix
            if self.rog:
                self.revision = self.revision[:-4]

    def _update_source(self) -> None:
        """Update the source of the firmware."""

        if isinstance(self.revision, int):
            self.source = FirmwareType.MERLIN
            return

        if isinstance(self.revision, str):
            if "gnuton" in self.revision:
                self.source = FirmwareType.GNUTON
                return
            if "alpha" in self.revision or "beta" in self.revision:
                self.source = FirmwareType.MERLIN
                return

        self.source = FirmwareType.STOCK

    def from_string(self, fw_string: Optional[str] = None) -> None:
        """Read the firmware string."""

        fw_string = clean_string(fw_string)
        if not fw_string:
            return None

        # Special cases for old firmwares and absent data
        if fw_string in ("__", "___"):
            return None

        pattern = (
            r"^(?P<major>[39].?0.?0.?[46])?[_.]?"
            r"(?P<minor>[0-9]{3})[_.]?"
            r"(?P<build>[0-9]+)[_.-]?"
            r"(?P<revision>[a-zA-Z0-9-_]+?)(?=_rog|$)?"
            r"(?P<rog>_rog)?$"
        )

        re_match = re.match(pattern, fw_string)
        if not re_match:
            if fw_string not in self._fw_warned:
                self._fw_warned.append(fw_string)
                _LOGGER.warning(
                    "Firmware version cannot be parsed. \
                        Please report this. The original FW string is: `%s`"
                    % fw_string
                )
            return None

        # Major version
        major = re_match.group("major")
        major = (
            major[0] + "." + major[1] + "." + major[2] + "." + major[3]
            if major and "." not in major and len(major) == 4
            else major
        )
        # Only if major version exists and has 0 member
        beta = major[0] == "9" if major and len(major) > 0 else False
        # Minor version
        minor = safe_int(re_match.group("minor"))
        # Build version
        build = safe_int(re_match.group("build"))
        # Revision
        revision = re_match.group("revision")
        revision = safe_int(revision) if revision.isdigit() else revision
        # ROG flag (Merlin firmware)
        rog = re_match.group("rog") == "_rog"

        self.major = major
        self.minor = minor
        self.build = build
        self.revision = revision
        self.rog = rog
        self.beta = beta

    def __str__(self) -> str:
        """Return the firmware version as a string."""

        return (
            f"{self.major}.{self.minor}.{self.build}_{self.revision}"
            + f"{'_rog' if self.rog else ''}"
        )

    def __repr__(self) -> str:
        """Return the firmware version as a string."""

        return self.__str__()

    def __eq__(self, other: object) -> bool:
        """Compare two firmware versions."""

        if not isinstance(other, Firmware):
            return False

        # Check if the major, minor, build and revision are the same
        # ROG flag for Merlin firmware is ignored / beta is in major
        return (
            self.major == other.major
            and self.minor == other.minor
            and self.build == other.build
            and self.revision == other.revision
        )

    def __lt__(self, other: object) -> bool:
        """Compare two firmware versions."""

        if not isinstance(other, Firmware):
            _LOGGER.debug("Cannot compare Firmware with other object")
            return False

        if self.source != other.source:
            _LOGGER.debug("Cannot compare different firmware sources")
            return False

        if self.beta != other.beta:
            _LOGGER.debug("Comparing beta and non-beta firmware")

        if not self.major or not other.major:
            return False

        # Remove beta digit from the major version
        major_1 = [int(x) for x in self.major.split(".")[1:]]
        major_2 = [int(x) for x in other.major.split(".")[1:]]

        def get_prefix(v: str) -> str:
            """Get the prefix of the version."""

            if "alpha" in v:
                return clean_string(v.split("alpha")[0]) or "-1"
            if "beta" in v:
                return clean_string(v.split("beta")[0]) or "-1"
            # We should not reach this point ever if
            # attributes of Firmware were not manually overwritten
            _LOGGER.warning(
                "Code execution reached an unexpected point"
                + " in Firmware.__lt__.get_prefix()."
                + " Trying to compare %s and %s" % (self, other)
            )
            return v

        def lt_values(v1: Any, v2: Any) -> bool:
            """Check whether one value is less than the other."""

            if v2 is None:
                return False
            elif v1 is None:
                return True

            if isinstance(v1, int) and isinstance(v2, int):
                return v1 < v2
            if isinstance(v1, str) and isinstance(v2, str):
                return v1 < v2
            if isinstance(v1, list) and isinstance(v2, list):
                for i in range(min(len(v1), len(v2))):
                    if v1[i] < v2[i]:
                        return True
                    if v2[i] < v1[i]:
                        return False
            # Case of beta / alpha
            if isinstance(v1, int) and isinstance(v2, str):
                _prefix = get_prefix(v2)
                _v1 = str(v1)
                # This case will solve `1 vs 1beta1`
                if _v1 == _prefix:
                    return False
                # Everything else can be compared as strings
                return str(v1) < _prefix
            if isinstance(v1, str) and isinstance(v2, int):
                _prefix = get_prefix(v1)
                _v2 = str(v2)
                # This case will solve `1 vs 1beta1`
                if _v2 == _prefix:
                    return True
                # Everything else can be compared as strings
                return _prefix < str(v2)

            return False

        # Compare the major version
        if lt_values(major_1, major_2):
            return True
        if lt_values(major_2, major_1):
            return False

        # Compare the minor version
        if lt_values(self.minor, other.minor):
            return True
        if lt_values(other.minor, self.minor):
            return False

        # Compare the build version
        if lt_values(self.build, other.build):
            return True
        if lt_values(other.build, self.build):
            return False

        # Compare the revision version
        if lt_values(self.revision, other.revision):
            return True
        if lt_values(other.revision, self.revision):
            return False

        return False

    def __gt__(self, other: object) -> bool:
        """Compare two firmware versions."""

        # Invert the statement of less-than
        return not self.__lt__(other) and self.__ne__(other)



"""Library init"""

from .asusrouter import AsusRouter
from .error import AsusRouterError
from .modules.data import AsusData
from .modules.endpoint import Endpoint
from .tools.dump import AsusRouterDump



"""Parental control module."""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Awaitable, Callable, Optional

from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.tools.converters import safe_int, safe_return

KEY_PC_BLOCK_ALL = "MULTIFILTER_BLOCK_ALL"
KEY_PC_MAC = "MULTIFILTER_MAC"
KEY_PC_NAME = "MULTIFILTER_DEVICENAME"
KEY_PC_STATE = "MULTIFILTER_ALL"
KEY_PC_TIMEMAP = "MULTIFILTER_MACFILTER_DAYTIME_V2"
KEY_PC_TYPE = "MULTIFILTER_ENABLE"

PC_RULE_MAP = {
    KEY_PC_MAC: "mac",
    KEY_PC_NAME: "name",
    KEY_PC_TIMEMAP: "timemap",
    KEY_PC_TYPE: "type",
}

HOOK_PC = [
    KEY_PC_BLOCK_ALL,
    KEY_PC_MAC,
    KEY_PC_NAME,
    KEY_PC_STATE,
    KEY_PC_TIMEMAP,
    KEY_PC_TYPE,
]


DEFAULT_PC_TIMEMAP = "W03E21000700<W04122000800"


class PCRuleType(IntEnum):
    """Parental control rule type."""

    UNKNOWN = -999
    REMOVE = -1  # pseudo type to remove a rule
    DISABLE = 0
    TIME = 1
    BLOCK = 2


@dataclass
class ParentalControlRule:
    """Parental control rule class."""

    mac: Optional[str] = None
    name: str = ""
    timemap: str = DEFAULT_PC_TIMEMAP
    type: PCRuleType = PCRuleType.UNKNOWN


class AsusParentalControl(IntEnum):
    """Asus parental control state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


class AsusBlockAll(IntEnum):
    """Asus block all state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusParentalControl | AsusBlockAll | ParentalControlRule,
    **kwargs: Any,
) -> bool:
    """Set the parental control state."""

    # Check if we need to set a rule
    if isinstance(state, ParentalControlRule):
        return await set_rule(callback, state, **kwargs)

    # Check if state is available and valid
    if not isinstance(
        state, (AsusParentalControl, AsusBlockAll)
    ) or not state.value in (0, 1):
        return False

    service_arguments = {}

    match state:
        case a if isinstance(a, AsusParentalControl):
            service_arguments = {
                KEY_PC_STATE: 1 if state == AsusParentalControl.ON else 0
            }
        case a if isinstance(a, AsusBlockAll):
            service_arguments = {KEY_PC_BLOCK_ALL: 1 if state == AsusBlockAll.ON else 0}

    # Get the correct service call
    service = "restart_firewall"

    # Call the service
    return await callback(
        service=service,
        arguments=service_arguments,
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )


async def set_rule(
    callback: Callable[..., Awaitable[bool]],
    rule: ParentalControlRule,
    **kwargs: Any,
) -> bool:
    """Set the parental control rule."""

    # Check if rule is available
    if not isinstance(rule, ParentalControlRule):
        return False

    # Get the current rules
    current_rules = (
        kwargs.get("router_state", {})
        .get(AsusData.PARENTAL_CONTROL, AsusDataState(data={}))
        .data.get("rules", {})
    )

    # Get rule action
    # If the rule is not available, we need to add it
    # update can also be handled as add
    action = add_rule
    if rule.type == PCRuleType.REMOVE:
        action = remove_rule

    # Perform the action
    current_rules = action(current_rules, rule)

    # Convert the rules to service arguments
    service_arguments = write_pc_rules(current_rules)

    # Get the correct service call
    service = "restart_firewall"

    # Call the service
    return await callback(
        service=service,
        arguments=service_arguments,
        apply=True,
    )


def check_rule(rule: Optional[ParentalControlRule]) -> Optional[ParentalControlRule]:
    """Check the parental control rule."""

    # Check if rule is available
    if not isinstance(rule, ParentalControlRule):
        return None

    # Check that mac is available
    if rule.mac is None:
        return None

    # Check that type is available and valid
    if not isinstance(rule.type, PCRuleType) or rule.type not in (
        PCRuleType.DISABLE,
        PCRuleType.TIME,
        PCRuleType.BLOCK,
    ):
        return None

    # Check that timemap is available and valid
    if not rule.timemap.strip():
        rule.timemap = DEFAULT_PC_TIMEMAP

    # Check that name is available
    if not rule.name.strip():
        rule.name = rule.mac

    # Return the rule
    return rule


def add_rule(
    current_rules: dict[str, ParentalControlRule],
    rule: Optional[ParentalControlRule] = None,
) -> dict[str, ParentalControlRule]:
    """Add a rule."""

    # Check that the current rules are available
    if not isinstance(current_rules, dict):
        current_rules = {}

    # Check if rule is available and valid
    rule = check_rule(rule)
    if rule is None or rule.mac is None:
        return current_rules

    # Add the new rule
    # This will also overwrite (update) the old rule if it exists
    current_rules[rule.mac] = rule

    return current_rules


def remove_rule(
    current_rules: dict[str, ParentalControlRule],
    rule: Optional[ParentalControlRule | str] = None,
) -> dict[str, ParentalControlRule]:
    """Remove a rule."""

    # Check that the current rules are available
    if not isinstance(current_rules, dict):
        current_rules = {}

    rule_mac = (
        rule.mac
        if isinstance(rule, ParentalControlRule)
        else rule
        if isinstance(rule, str)
        else None
    )

    # If mac is available, remove the rule
    if rule_mac is not None:
        current_rules.pop(rule_mac, None)

    return current_rules


def read_pc_string(key: str, data: dict[str, str]) -> list[str]:
    """Read the parental control string."""

    return data.get(key, "").split("&#62")


def read_pc_rules(data: dict[str, str]) -> dict[str, ParentalControlRule]:
    """Read the parental control data."""

    # If no data is provided, return empty list
    if data.get(KEY_PC_MAC) == data.get(KEY_PC_TYPE):
        return {}

    # The data is split in 4 strings. Each data value is split in the string
    # with a `&#62` separator. We need to map the data and make sure, that
    # each `ParentalControlRule` has all values

    # Map the values to a list of `ParentalControlRule`
    rules = {}
    for rule_mac, rule_name, rule_timemap, rule_type in zip(
        *[read_pc_string(key, data) for key in PC_RULE_MAP]
    ):
        # Map the values
        rule = ParentalControlRule(
            mac=safe_return(rule_mac),
            name=safe_return(rule_name),
            timemap=safe_return(rule_timemap),
            type=PCRuleType(safe_int(rule_type, default=-999)),
        )

        # Append the rule to the list
        rules[rule_mac] = rule

    return rules


def write_pc_rules(rules: dict[str, ParentalControlRule]) -> dict[str, str]:
    """Write the parental control data."""

    # If no rules are provided, return empty dict
    if not rules:
        return {key: "" for key in PC_RULE_MAP}

    # Join the values together
    data = {}
    for key, attribute in PC_RULE_MAP.items():
        data[key] = ">".join(
            str(getattr(rule, attribute, "")) for rule in rules.values()
        )

    data[KEY_PC_TIMEMAP] = data[KEY_PC_TIMEMAP].replace("&#60", "<")

    return data



"""Port Status endpoint module."""

from __future__ import annotations

import logging
from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.modules.ports import (
    PORT_CAP2TYPE,
    PORT_TYPE2LINK,
    PortCapability,
    PortSpeed,
    PortType,
)
from asusrouter.tools.converters import (
    get_enum_key_by_value,
    int_as_capabilities,
    safe_bool,
    safe_int,
)
from asusrouter.tools.readers import read_json_content as read  # noqa: F401

_LOGGER = logging.getLogger(__name__)


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process port status data"""

    ports: dict[PortType, Any] = {}

    if not data:
        return {
            AsusData.PORTS: ports,
            AsusData.NODE_INFO: {},
        }

    state: dict[AsusData, Any] = {}

    # Node info
    state[AsusData.NODE_INFO] = process_node_info(data)

    # Port info
    port_info = data.get("port_info", {})
    if port_info:
        for mac, info in port_info.items():
            ports[mac] = {}

            for port, values in info.items():
                # Process the port info
                port_description, port_type, port_id = process_port_info(
                    port, values
                )

                # Create a port type group if it doesn't exist yet
                if port_type not in ports[mac]:
                    ports[mac][port_type] = {}

                # Save the port info
                ports[mac][port_type][port_id] = port_description

    state[AsusData.PORTS] = ports

    return state


def process_node_info(data: dict[str, Any]) -> dict[str, Any]:
    """Process node info data."""

    node_info = data.get("node_info", {})

    if not node_info:
        return {}

    nodes = {}

    for mac, info in node_info.items():
        # Save the node info
        nodes[mac] = info

    return nodes


def process_port_info(
    port: str, values: dict[str, Any]
) -> tuple[dict[str, Any], PortType, int]:
    """Process port info data."""

    # The port is a string with the format `port_label:port_id`
    # e.g. `L1` for LAN port 1 or `W0` for WAN port 0 (the main)
    # port_label = port[0]
    port_id = safe_int(port[1])

    # Get the capabilities of the port
    port_capabilities = int_as_capabilities(
        safe_int(values.get("cap")), PortCapability
    )

    # Get the port type
    port_type = PortType.UNKNOWN
    for pcap, ptype in PORT_CAP2TYPE.items():
        if port_capabilities.get(pcap) is True:
            port_type = ptype
            break

    # Get the correct link enum
    link_enum: Any = PORT_TYPE2LINK.get(port_type)

    # Get the rates. Set to 0 if not available
    link_rate = safe_int(values.get("link_rate"), default=0)
    max_rate = safe_int(values.get("max_rate"), default=0)

    if link_enum is not None:
        link_rate = get_enum_key_by_value(
            link_enum,
            link_rate,
            link_enum.UNKNOWN,
        )
        max_rate = get_enum_key_by_value(
            link_enum,
            max_rate,
            link_enum.UNKNOWN,
        )

    # Special ports
    if max_rate == PortSpeed.LINK_10000:
        if port_capabilities.get(PortCapability.SFPP) is True:
            port_type = PortType.SFPP

    # Port state
    port_state = safe_bool(values.get("is_on"))
    # For USB ports, the state is 1 only when a modem is connected
    modem = False
    port_devices = None
    if port_type == PortType.USB:
        # Mark a modem as connected
        if port_state is True:
            modem = True
        # If any other device is connected, mark as connected
        port_devices = values.get("devices")
        if port_devices is not None:
            port_state = True

    # Leave only the capabilities that are available
    capabilities = [
        capability
        for capability, value in port_capabilities.items()
        if value is True
    ]

    # Combine the port description
    port_description = {
        "state": port_state,
        "id": port_id,
        "capabilities": capabilities,
        "link_rate": link_rate,
        "max_rate": max_rate,
    }

    # Add the modem flag if needed and the connected devices
    if port_type == PortType.USB:
        port_description["modem"] = modem
        port_description["devices"] = port_devices

    return port_description, port_type, port_id



"""Constants for the devicemap endpoint module."""

# These values are just stored directly in this order in the corresponding node
# Format: (output_group, input_group, [input_values])
DEVICEMAP_BY_INDEX = [
    ("wan", "wan", ["status", "sbstatus", "auxstatus"]),
    ("wan0", "first_wan", ["status", "sbstatus", "auxstatus"]),
    ("wan1", "second_wan", ["status", "sbstatus", "auxstatus"]),
    ("usb", "usb", ["status"]),
]

# These values are stored as "key=value"
# Format: (output_group, input_group, [input_values])
DEVICEMAP_BY_KEY = [
    (
        "wan",
        "wan",
        [
            "monoClient",
            "wlc_state",
            "wlc_sbstate",
            "wifi_hw_switch",
            "ddnsRet",
            "ddnsUpdate",
            "wan_line_state",
            "wlan0_radio_flag",
            "wlan1_radio_flag",
            "wlan2_radio_flag",
            "data_rate_info_2g",
            "data_rate_info_5g",
            "data_rate_info_5g_2",
            "wan_diag_state",
            "active_wan_unit",
            "wlc0_state",
            "wlc1_state",
            "rssi_2g",
            "rssi_5g",
            "rssi_5g_2",
            "link_internet",
            "wlc2_state",
            "le_restart_httpd",
        ],
    ),
    (
        "vpn",
        "vpn",
        [
            "vpnc_proto",
            "vpnc_state_t",
            "vpnc_sbstate_t",
            "vpn_client1_state",
            "vpn_client2_state",
            "vpn_client3_state",
            "vpn_client4_state",
            "vpn_client5_state",
            "vpnd_state",
            "vpn_client1_errno",
            "vpn_client2_errno",
            "vpn_client3_errno",
            "vpn_client4_errno",
            "vpn_client5_errno",
            "vpn_server1_state",
            "vpn_server2_state",
        ],
    ),
    (
        "sys",
        "sys",
        [
            "uptimeStr",
        ],
    ),
    (
        "qtn",
        "qtn",
        [
            "qtn_state",
        ],
    ),
    (
        "usb",
        "usb",
        [
            "modem_enable",
        ],
    ),
    (
        "wan0",
        "wan",
        [
            "wan0_enable",
            "wan0_realip_state",
            "wan0_ipaddr",
            "wan0_realip_ip",
        ],
    ),
    (
        "wan1",
        "wan",
        [
            "wan1_enable",
            "wan1_realip_state",
            "wan1_ipaddr",
            "wan1_realip_ip",
        ],
    ),
    (
        "sim",
        "sim",
        [
            "sim_state",
            "sim_signal",
            "sim_operation",
            "sim_isp",
            "roaming",
            "roaming_imsi",
            "sim_imsi",
            "g3err_pin",
            "pin_remaining_count",
            "modem_act_provider",
            "rx_bytes",
            "tx_bytes",
            "modem_sim_order",
        ],
    ),
    (
        "dhcp",
        "dhcp",
        [
            "dnsqmode",
        ],
    ),
    (
        "diag",
        "diag",
        [
            "diag_dblog_enable",
            "diag_dblog_remaining",
        ],
    ),
]

# This data should be cleared from some symbols
# Format: {output_group: {key: value_to_remove}}
DEVICEMAP_CLEAR = {
    "wan": {
        "data_rate_info_2g": '"',
        "data_rate_info_5g": '"',
        "data_rate_info_5g_2": '"',
    }
}

# This data is stored in a special way
# psta:wlc_state=0;wlc_state_auth=0;
# as a single value
# We need to split it into multiple values and store them in the corresponding group
# Format: (output_group, input_group, input_subgroup, [input_values])
# This part is not used in the current version of the library
DEVICEMAP_SPECIAL = [
    ("wan", "wan", "psta", ["wlc_state", "wlc_state_auth"]),
]



"""SysInfo endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.modules.wlan import WLAN_TYPE, Wlan
from asusrouter.tools.cleaners import clean_content
from asusrouter.tools.converters import safe_float, safe_int
from asusrouter.tools.readers import read_json_content


def read(content: str) -> dict[str, Any]:
    """Read sysinfo data"""

    # Prepare the content
    content = clean_content(content).replace(" = ", '":').replace(";\n", ',"')
    content = '{"' + content[:-3] + "}"

    # Read the json content
    sysinfo: dict[str, Any] = read_json_content(content)

    return sysinfo


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process sysinfo data"""

    state: dict[AsusData, Any] = {}

    sysinfo = {}

    # WLAN info
    wlan_info = {}
    i = 0
    while wlan_data := data.get(f"wlc_{i}_arr"):
        name = WLAN_TYPE.get(i, Wlan.UNKNOWN)
        wlan_info[name] = {
            "client_associated": safe_int(wlan_data[0]),
            "client_authorized": safe_int(wlan_data[1]),
            "client_authenticated": safe_int(wlan_data[2]),
        }
        i += 1
    sysinfo["wlan"] = wlan_info

    # Connections info
    connections_info = {}
    connections_data = data.get("conn_stats_arr")
    if connections_data:
        connections_info = {
            "total": safe_int(connections_data[0]),
            "active": safe_int(connections_data[1]),
        }
    sysinfo["connections"] = connections_info

    # Memory info
    memory_info = {}
    memory_data = data.get("mem_stats_arr")
    if memory_data:
        # Before 388.7
        # JFFS data is presented as a string of `XX.xx / YY.yy MB`
        # where `XX.xx` is the used space (float) and `YY.yy` is the total space (float)
        jffs = memory_data[7]
        if "/" in jffs:
            jffs_data = jffs[:-3].split(" / ")
            jffs_used = safe_float(jffs_data[0])
            jffs_total = safe_float(jffs_data[1])
            jffs_free = jffs_total - jffs_used if jffs_used and jffs_total else None
        # From 388.7
        # JFFS is just a `free` single float
        else:
            jffs_free = safe_float(jffs)
            jffs_used = None
            jffs_total = None

        memory_info = {
            "total": safe_float(memory_data[0]),
            "free": safe_float(memory_data[1]),
            "buffers": safe_float(memory_data[2]),
            "cache": safe_float(memory_data[3]),
            "swap_1": safe_float(memory_data[4]),
            "swap_2": safe_float(memory_data[5]),
            "nvram": safe_int(memory_data[6]),
            "jffs_free": jffs_free,
            "jffs_used": jffs_used,
            "jffs_total": jffs_total,
        }
    sysinfo["memory"] = memory_info

    # Load average info
    load_avg_info = {}
    load_avg_data = data.get("cpu_stats_arr")
    if load_avg_data:
        load_avg_info = {
            1: safe_float(load_avg_data[0]),
            5: safe_float(load_avg_data[1]),
            15: safe_float(load_avg_data[2]),
        }
    sysinfo["load_avg"] = load_avg_info

    # Sysinfo as it is
    state[AsusData.SYSINFO] = sysinfo

    return state



"""VPN endpoint module."""

from __future__ import annotations

import re
from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.modules.openvpn import AsusOVPNClient, AsusOVPNServer
from asusrouter.tools.cleaners import clean_content
from asusrouter.tools.converters import safe_int
from asusrouter.tools.readers import read_as_snake_case, read_js_variables

from .vpn_const import MAP_OVPN_CLIENT, MAP_OVPN_SERVER

_TO_IGNORE = ("", "None", "0.0.0.0")


def read(content: str) -> dict[str, Any]:
    """Read VPN data."""

    vpn: dict[str, Any] = {
        "client": {},
        "server": {},
    }

    # Read the js variables
    values: dict[str, Any] = read_js_variables(clean_content(content))

    # Values to ignore

    # Read the OpenVPN data
    for party in ("client", "server"):
        # Select proper method and state
        method = read_ovpn_client if party == "client" else read_ovpn_server
        state = AsusOVPNClient if party == "client" else AsusOVPNServer

        for num in range(1, 11):
            key = f"vpn_{party}{num}_status"
            if key not in values:
                break

            # Read the data
            if values[key] not in _TO_IGNORE:
                vpn[party][num] = method(values[key])
                vpn[party][num]["state"] = state.CONNECTED
            else:
                vpn[party][num] = {"state": state.DISCONNECTED}

            # Read additional data
            for add_key in ("ip", "rip"):
                search_value = values.get(f"vpn_{party}{num}_{add_key}")
                if search_value and search_value not in _TO_IGNORE:
                    vpn[party][num][add_key] = search_value

            # Read remote
            search_value = vpn[party][num].get("remote")
            if search_value:
                vpn[party][num]["remote"] = read_ovpn_remote(search_value)

    return vpn


def read_ovpn_client(content: str) -> dict[str, Any]:
    """Read OpenVPN client data."""

    ovpn_client: dict[str, Any] = {}

    # Strip the content and make sure it's not an empty string
    content = content.strip()
    if content in ("", "None"):
        return ovpn_client

    # --------------------
    # THIS PART IS LEGACY AND CAN BE REMOVED IN THE FUTURE -->
    # --------------------

    for value in MAP_OVPN_CLIENT:
        key_old, key_new, method = value
        value = re.search(f"{key_old},(.*?)(?=>)", content)
        if value:
            ovpn_client[key_new] = method(value[1]) if method else value[1]

    # --------------------
    # <-- THIS PART IS LEGACY AND CAN BE REMOVED IN THE FUTURE
    # --------------------

    return ovpn_client


def read_ovpn_remote(content: str) -> dict[str, Any]:
    """Read OpenVPN remote data."""

    try:
        parts = content.split(",")
        auth = parts[1]
        parts = parts[0].split(":")
        ip_address = parts[0]
        port = safe_int(parts[1])
    except IndexError:
        return {}

    return {
        "ip": ip_address,
        "port": port,
        "auth": auth,
    }


def read_ovpn_server(content: str) -> dict[str, Any]:
    """Read OpenVPN server data."""

    ovpn_server: dict[str, Any] = {}

    # Strip the content and make sure it's not an empty string
    content = content.strip()
    if content in ("", "None"):
        return ovpn_server

    # --------------------
    # THIS PART IS LEGACY AND CAN BE REMOVED IN THE FUTURE -->
    # --------------------

    # Map the values
    for value in MAP_OVPN_SERVER:
        key_old, key_new = value
        if key_old in content:
            flag = False
            keys = []
            # Find headers
            header = re.search(f"HEADER,{key_old},(.*?)(?=>)", content)
            if header:
                keys_raw = header[1].split(",")
                for element in keys_raw:
                    keys.append(read_as_snake_case(element))
                flag = True

            # Hide header
            content = content.replace(f"HEADER,{key_old},", "USED-")

            ovpn_server[key_new] = []
            # Find values
            while value := re.search(f"{key_old},(.*?)(?=>)", content):
                if not value:
                    break
                if flag:
                    cut = value[1].split(",")
                    array = {k: v for k, v in zip(keys, cut)}
                    ovpn_server[key_new].append(array)
                else:
                    ovpn_server[key_new].append(value[1])
                content = content.replace(f"{key_old},{value[1]}", "USED-")
    # --------------------
    # <-- THIS PART IS LEGACY AND CAN BE REMOVED IN THE FUTURE
    # --------------------

    return ovpn_server


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process VPN data."""

    vpn: dict[AsusData, Any] = {
        AsusData.OPENVPN_CLIENT: data.get("client"),
        AsusData.OPENVPN_SERVER: data.get("server"),
    }

    return vpn



"""Starting point for the AsusRouter."""

from __future__ import annotations

import argparse
import asyncio
import logging

import aiohttp

from asusrouter import AsusRouter, AsusRouterDump, AsusRouterError
from asusrouter.modules.data import AsusData

_LOGGER = logging.getLogger(__name__)


async def _connect_and_dump(args: argparse.Namespace) -> None:
    """Connect to the router and dump the data."""

    _LOGGER.info("Starting dump...")

    with AsusRouterDump(args.output, args.dump, args.zip) as dump:
        async with aiohttp.ClientSession() as session:
            _LOGGER.debug("Created aiohttp session")

            # Create the router
            router = AsusRouter(
                hostname=args.host,
                username=args.username,
                password=args.password,
                use_ssl=args.ssl,
                session=session,
                dumpback=dump.dump,
            )
            _LOGGER.debug("Created router")

            # Connect to the router and gather identity data
            _LOGGER.debug("Connecting to the router...")
            await router.async_connect()

            if not router.connected:
                _LOGGER.error("Failed to connect to the router")
                return

            _LOGGER.debug("Connected and identified")

            _LOGGER.debug("Checking all known data...")

            for datatype in AsusData:
                await router.async_get_data(datatype)

            _LOGGER.debug("Finished checking all known data")

            # Disconnect from the router
            await router.async_disconnect()
            _LOGGER.debug("Disconnected")

    _LOGGER.info("Dump finished. Saved to: %s", args.output)


def main():
    """Run AsusRouter as a program."""

    parser = argparse.ArgumentParser(
        description="AsusRouter package command line interface."
    )
    parser.add_argument(
        "-o", "--output", type=str, required=True, help="The output folder for the log."
    )
    parser.add_argument("--dump", action="store_true", help="Perform a full dump.")
    parser.add_argument(
        "--host",
        type=str,
        required=True,
        help="The hostname / ip address of the router.",
    )
    parser.add_argument(
        "-u",
        "--username",
        type=str,
        default="admin",
        help="The username for the router. Default: `admin`.",
    )
    parser.add_argument(
        "-p",
        "--password",
        type=str,
        default="admin",
        help="The password for the router. Default: `admin`.",
    )
    parser.add_argument(
        "-s",
        "--ssl",
        action="store_true",
        default=False,
        help="Use SSL for the connection.",
    )
    parser.add_argument(
        "-P",
        "--port",
        type=int,
        default=None,
        help="The port to connect to. Only when using non-default settings.\
            Default is 80 for HTTP and 443 for HTTPS.",
    )
    parser.add_argument("--debug", action="store_true", help="Enable debug logging.")
    parser.add_argument(
        "--zip", action="store_true", default=False, help="Zip the output."
    )

    args = parser.parse_args()

    # Set up logging
    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format="%(asctime)s [%(levelname)s] %(module)s: %(message)s",
    )

    # Connect to the router and dump the data
    try:
        asyncio.run(_connect_and_dump(args))
    except AsusRouterError as ex:
        _LOGGER.error("AsusRouterError: %s", ex)


if __name__ == "__main__":
    main()



"""Constants for the VPN endpoint module."""

from asusrouter.tools.converters import safe_datetime, safe_int

MAP_OVPN_CLIENT = [
    ("REMOTE", "remote", None),
    ("Updated", "datetime", safe_datetime),
    ("TUN/TAP read bytes", "tun_tap_read", safe_int),
    ("TUN/TAP write bytes", "tun_tap_write", safe_int),
    ("TCP/UDP read bytes", "tcp_udp_read", safe_int),
    ("TCP/UDP write bytes", "tcp_udp_write", safe_int),
    ("Auth read bytes", "auth_read", safe_int),
    ("pre-compress bytes", "pre_compress", safe_int),
    ("post-compress bytes", "post_compress", safe_int),
    ("pre-decompress bytes", "pre_decompress", safe_int),
    ("post-decompress bytes", "post_decompress", safe_int),
]

MAP_OVPN_SERVER = [
    ("CLIENT_LIST", "client_list"),
    ("ROUTING_TABLE", "routing_table"),
]



"""AiMesh module."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class AiMeshDevice:
    """AiMesh device class"""

    # Status
    status: bool = False

    alias: str | None = None
    model: str | None = None
    product_id: str | None = None
    ip: str | None = None

    fw: str | None = None
    fw_new: str | None = None

    mac: str | None = None

    # Access point: ap2g, ap5g, ap5g1, ap6g, apdwb
    ap: dict[str, Any] | None = None
    # Parent AiMesh: pap2g, rssi2g, pap2g_ssid, pap5g, rssi5g, pap5g_ssid, pap6g, rssi6g, pap6g_ssid
    parent: dict[str, Any] | None = None
    # Node state
    type: str | None = None
    level: int | None = None
    config: dict[str, Any] | None = None



"""Cleaning tools for AsusRouter."""

from __future__ import annotations

from typing import Any


def clean_content(content: str) -> str:
    """Clean the content from useless data"""

    # Remove the first character if it is a BOM
    if content.startswith("\ufeff"):
        content = content[1:]

    # Return the cleaned content
    return content


def clean_dict(data: dict[str, Any]) -> dict[str, Any]:
    """Clean a dict from useless data

    This method performs recursively (to all the nested dicts) the following operations:
    - Convert all values which are empty strings ('') to None"""

    # Go through the dict and clean it
    for key, value in data.items():
        # Check if the value is a dict
        if isinstance(value, dict):
            # Clean the dict
            data[key] = clean_dict(value)
            # Continue to the next value
            continue
        # Check if the value is a string and if it is empty
        if isinstance(value, str) and value == "":
            # Convert the value to None
            data[key] = None

    # Return the cleaned dict
    return data


def clean_dict_key_prefix(data: dict[str, Any], prefix: str) -> dict[str, Any]:
    """Clean dict keys from prefix and an underscore (prefix_).

    This method only cleans the keys, but cannot be used to clean nested dicts."""

    cleaned_dict: dict[str, Any] = {}

    # Go through the dict and clean it
    for key, value in data.items():
        # Check if the key starts with the prefix
        if key.startswith(f"{prefix}_"):
            # Remove the prefix and underscore
            new_key = key.replace(f"{prefix}_", "")
            # Add the new key to the dict
            cleaned_dict[new_key] = value

            continue
        cleaned_dict[key] = value

    # Return the cleaned dict
    return cleaned_dict


def clean_dict_key(data: dict[str, Any], keys: str | list[str]) -> dict[str, Any]:
    """Clean dict from the keys. This method can be used to clean nested dicts."""

    cleaned_dict: dict[str, Any] = {}

    # Check if the keys is a string
    if isinstance(keys, str):
        keys = [keys]

    # Go through the dict and clean it
    for key, value in data.items():
        # Check if the key is in the keys
        if key in keys:
            continue
        cleaned_dict[key] = (
            clean_dict_key(value, keys) if isinstance(value, dict) else value
        )

    # Return the cleaned dict
    return cleaned_dict



"""Vendor module."""


def replace_vendor(vendor: str) -> str:
    """Replace vendor name."""

    # Check if vendor is any of `android-dhcp-XX` where XX is a version number
    if vendor.startswith("android-dhcp-"):
        # Get the version number
        version = vendor.replace("android-dhcp-", "")
        # Return the version number
        return f"Android {version}"

    match vendor:
        case "MSFT 5.0":
            return "Microsoft Corporation"

    return vendor



"""Ethernet ports endpoint module."""

from __future__ import annotations

import logging
from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.modules.ports import PORT_SPEED, PortSpeed, PortType
from asusrouter.tools.converters import safe_int
from asusrouter.tools.readers import read_json_content

_LOGGER = logging.getLogger(__name__)


def read(content: str) -> dict[str, Any]:
    """Read ethernet ports data."""

    # Read the json content
    ethernet_ports: dict[str, Any] = read_json_content(
        content.replace("get_wan_lan_status = ", "").replace(";", "")
    )

    return ethernet_ports


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process ethernet ports data."""

    # Ports info
    ports: dict[PortType, dict] = {
        PortType.LAN: {},
        PortType.WAN: {},
    }

    port_speed = data.get("portSpeed")

    if port_speed is None:
        return {
            AsusData.PORTS: ports,
        }

    for port, value in port_speed.items():
        # Get the port code
        port_code = port[0:3].lower()
        # Check whether the port code is in PortType enum
        try:
            port_type = PortType(port_code)
        except ValueError:
            # This should be some other kind of port and not LAN or WAN
            # Based on https://github.com/Vaskivskyi/ha-asusrouter/issues/774
            # it is probably the SFPP port, since 10G WAN/LAN should be
            # detected properly.
            if port_code == "10g":
                port_type = PortType.SFPP
            else:
                continue

            # If the port type is not in the ports dict, add it
            if port_type not in ports:
                ports[port_type] = {}

        # Get the port id and link rate
        port_id = safe_int(port[3:])
        link_rate = PORT_SPEED.get(value)
        # Save the port info
        ports[port_type][port_id] = {
            "state": link_rate not in (PortSpeed.LINK_DOWN, PortSpeed.UNKNOWN),
            "link_rate": link_rate,
        }

    return {
        AsusData.PORTS: ports,
    }



"""RGB endpoint module."""

from __future__ import annotations

from typing import Any


def read(content: str) -> dict[str, Any]:  # pylint: disable=unused-argument
    """Read RGB data"""

    # Read the page content
    rgb: dict[str, Any] = {}

    return rgb



"""Network tools endpoint module."""

from __future__ import annotations

from typing import Any

from asusrouter.modules.data import AsusData
from asusrouter.tools.readers import read_json_content as read  # noqa: F401


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process network tools data."""

    return {
        AsusData.PING: data,
    }



"""Data transform module."""

from __future__ import annotations

from typing import Any, Optional

from asusrouter.modules.client import process_client
from asusrouter.modules.data import AsusDataState
from asusrouter.modules.ports import PortType
from asusrouter.tools.readers import readable_mac

# List of models with 6Ghz support
# and no 5Ghz2 support
MODEL_WITH_6GHZ = [
    "RT-AXE95Q",
]


def transform_network(
    data: dict[str, Any],
    services: Optional[list[str]],
    history: Optional[AsusDataState],
    **kwargs: Any,
) -> dict[str, Any]:
    """Transform network data."""

    # Check if the device has dualwan support
    if not services:
        return data
    if not "dualwan" in services:
        return data

    network = data.copy()
    # Add speed if not available - fix first empty value round
    for interface in network:
        for speed in ("rx_speed", "tx_speed"):
            if not speed in network[interface]:
                network[interface][speed] = 0.0

    # Add usb network if not available
    if not "usb" in network:
        # Check history
        usb_history = history.data.get("usb") if history and history.data else None
        # Revert to history if available
        if usb_history:
            network["usb"] = usb_history
            network["usb"]["rx_speed"] = 0.0
            network["usb"]["tx_speed"] = 0.0
        else:
            network["usb"] = {
                "rx": 0,
                "tx": 0,
                "rx_speed": 0.0,
                "tx_speed": 0.0,
            }

    # Get the model if available
    model = kwargs.get("model", None)

    # Check if we have 5GHz2 available in the network data
    if "5ghz2" in network:
        # Check interfaces for 5Ghz2/6Ghz
        support_5ghz2 = "5G-2" in services
        support_6ghz = "wifi6e" in services

        if (
            support_5ghz2 is False and support_6ghz is True
        ) or model in MODEL_WITH_6GHZ:
            # Rename 5Ghz2 to 6Ghz
            network["6ghz"] = network.pop("5ghz2")

    return network


def transform_clients(
    data: dict[str, Any], history: Optional[AsusDataState], **kwargs: Any
) -> dict[str, Any]:
    """Transform clients data."""

    clients = {}
    for mac, client in data.items():
        if readable_mac(mac):
            # Check client history
            client_history = history.data.get(mac) if history and history.data else None
            # Process the client
            clients[mac] = process_client(client, client_history, **kwargs)

    return clients


def transform_cpu(
    data: dict[str, dict[str, Any]],
) -> dict[str, Any]:
    """Transform cpu data."""

    for info in data.values():
        info.setdefault("usage", None)

    return data


def transform_ethernet_ports(
    data: dict[str, Any],
    mac: Optional[str],
) -> dict[str, dict[str, Any]]:
    """Transform the legacy ethernet ports data to the new format."""

    # Check if the first level of the dict is PortType enum
    # If any other key is found, return the data as is
    for data_key in data:
        if not isinstance(data_key, PortType):
            return data

    # If mac is not available, return the data as is
    if not mac:
        return data

    # Transform the data
    return {mac: data}


def transform_wan(
    data: dict[str, Any],
    services: Optional[list[str]],
) -> dict[str, Any]:
    """Transform WAN data."""

    wan = data.copy()

    if not services:
        return wan

    service_keys = {"dualwan": "dualwan", "wanbonding": "aggregation"}

    for service, key in service_keys.items():
        if service not in services:
            wan.pop(key, None)

    return wan



"""Devicemap endpoint module."""

from __future__ import annotations

import logging
import re
from datetime import datetime, timedelta
from typing import Any, Optional, Tuple

import xmltodict

from asusrouter.modules.data import AsusData, AsusDataState
from asusrouter.modules.endpoint import data_get
from asusrouter.modules.openvpn import AsusOVPNClient, AsusOVPNServer
from asusrouter.tools.cleaners import clean_dict, clean_dict_key_prefix
from asusrouter.tools.converters import safe_datetime, safe_int
from asusrouter.tools.readers import merge_dicts

from .devicemap_const import (
    DEVICEMAP_BY_INDEX,
    DEVICEMAP_BY_KEY,
    DEVICEMAP_CLEAR,
)

_LOGGER = logging.getLogger(__name__)

REQUIRE_HISTORY = True


def read(content: str) -> dict[str, Any]:
    """Read devicemap data."""

    # Create a dict to store the data
    devicemap: dict[str, Any] = {}

    # Parse the XML data
    try:
        xml_content: dict[str, Any] = xmltodict.parse(content).get(
            "devicemap", {}
        )
        if not xml_content:
            _LOGGER.debug("Received empty devicemap XML")
            return devicemap
    except xmltodict.expat.ExpatError as ex:  # type: ignore
        _LOGGER.debug("Received invalid devicemap XML: %s", ex)
        return devicemap

    # Go through the data and fill the dict

    # Get values by index using read_devicemap_index method
    devicemap = merge_dicts(devicemap, read_index(xml_content))

    # Get values by key using read_devicemap_key method
    devicemap = merge_dicts(devicemap, read_key(xml_content))

    # Clear values from useless symbols
    for output_group, clear_map in DEVICEMAP_CLEAR.items():
        if output_group not in devicemap:
            continue
        for key, clear_value in clear_map.items():
            # If the key is not in the devicemap, continue
            if key not in devicemap[output_group]:
                continue
            devicemap[output_group][key] = devicemap[output_group][
                key
            ].replace(clear_value, "")

    # Clean the devicemap values
    devicemap = clean_dict(devicemap)

    # Clean the devicemap values_dict from output_group prefix
    for output_group, values_dict in devicemap.items():
        devicemap[output_group] = clean_dict_key_prefix(
            values_dict, output_group
        )

    # Return the devicemap
    return devicemap


def read_index(xml_content: dict[str, Any]) -> dict[str, Any]:
    """Read devicemap by index.

    This method performs reading of the devicemap by index
    to simplify the original read_devicemap method."""

    # Create a dict to store the data
    devicemap: dict[str, Any] = {}

    # Get values for which we only know their order (index)
    for output_group, input_group, input_values in DEVICEMAP_BY_INDEX:
        # Create an empty dictionary for the output group
        devicemap[output_group] = {}

        # Check that the input group is in the xml content
        if input_group not in xml_content:
            continue

        # Use dict comprehension to build output_group_data
        output_group_data = {
            input_value: xml_content[input_group][index]
            for index, input_value in enumerate(input_values)
            if index < len(xml_content[input_group])
        }

        # Add the output group data to the devicemap
        devicemap[output_group] = output_group_data

    # Return the devicemap
    return devicemap


def read_key(xml_content: dict[str, Any]) -> dict[str, Any]:
    """Read devicemap by key.

    This method performs reading of the devicemap by key
    to simplify the original read_devicemap method."""

    # Create a dict to store the data
    devicemap: dict[str, Any] = {}

    # Get values for which we know their key
    for output_group, input_group, input_values in DEVICEMAP_BY_KEY:
        # Create a dict to store the data
        output_group_data: dict[str, Any] = {}

        # Go through the input values and fill the dict
        for input_value in input_values:
            # Get the input group data
            xml_input_group = xml_content.get(input_group)

            # If the input group data is None, skip this iteration
            if xml_input_group is None:
                continue

            # If the input group data is a string, convert it to a list
            if isinstance(xml_input_group, str):
                xml_input_group = [xml_input_group]

            # Go through the input group data and check if the input value is in it
            for value in xml_input_group:
                if input_value in value:
                    # Add the input value to the output group data
                    output_group_data[input_value] = value.replace(
                        f"{input_value}=", ""
                    )
                    break

        # Add the output group data to the devicemap
        devicemap[output_group] = output_group_data

    # Return the devicemap
    return devicemap


# This method performs reading of the devicemap special values
# pylint: disable-next=unused-argument
def read_special(xml_content: dict[str, Any]) -> dict[str, Any]:
    """Read devicemap special values."""

    # This method is not implemented yet

    return {}


def read_uptime_string(
    content: str,
) -> Tuple[Optional[datetime], Optional[int]]:
    """Read uptime string and return proper datetime object."""

    # Split the content into the date/time part and the seconds part
    uptime_parts = content.split("(")
    if len(uptime_parts) < 2:
        return (None, None)

    # Extract the number of seconds from the seconds part
    seconds_match = re.search("([0-9]+)", uptime_parts[1])
    if not seconds_match:
        return (None, None)
    seconds = safe_int(seconds_match.group())

    when = safe_datetime(uptime_parts[0])
    if when is None:
        return (None, seconds)

    # This part will always work, since seconds are always an integer
    # unless `safe_int` got broken
    uptime = when - timedelta(seconds=seconds)
    return (uptime, seconds)


def process(data: dict[str, Any]) -> dict[AsusData, Any]:
    """Process data from devicemap endpoint."""

    # Get the passed awrguments
    history: dict[AsusData, AsusDataState] = data_get(data, "history") or {}

    # Devicemap - just the data itself
    devicemap = data

    # Boot time
    prev_boottime_object: Optional[AsusDataState] = history.get(
        AsusData.BOOTTIME
    )
    prev_boottime = prev_boottime_object.data if prev_boottime_object else None
    boottime, reboot = process_boottime(devicemap, prev_boottime)

    # Mark reboot
    flags = {}
    if reboot:
        flags["reboot"] = True

    # OpenVPN
    openvpn = process_ovpn(devicemap)

    # Return the processed data
    return {
        AsusData.DEVICEMAP: devicemap,
        AsusData.BOOTTIME: boottime,
        AsusData.OPENVPN: openvpn,
        AsusData.FLAGS: flags,
    }


def process_boottime(
    devicemap: dict[str, Any], prev_boottime: Optional[dict[str, Any]]
) -> Tuple[dict[str, Any], bool]:
    """Process boottime data"""

    # Reboot flag
    reboot = False

    boottime: dict[str, Any] = {}

    # Since precision is 1 second, could be that old and new
    # are 1 sec different. In this case, we should not change
    # the boot time, but keep the previous value to avoid regular changes
    sys = devicemap.get("sys")
    if sys:
        uptime_str = sys.get("uptimeStr")
        _LOGGER.debug("Uptime string: %s", uptime_str)
        if uptime_str:
            time, seconds = read_uptime_string(uptime_str)
            if time:
                boottime["datetime"] = time
                if prev_boottime and "datetime" in prev_boottime:
                    delta = time - prev_boottime["datetime"]

                    # Check for reboot
                    if abs(delta.seconds) >= 2 and delta.seconds >= 0:
                        reboot = True
                    else:
                        boottime = prev_boottime

                boottime["uptime"] = seconds

    return boottime, reboot


def process_ovpn(devicemap: dict[str, Any]) -> dict[str, Any]:
    """Process OpenVPN data"""

    vpn: dict[str, Any] = {
        "client": {},
        "server": {},
    }
    vpnmap = devicemap.get("vpn")

    if vpnmap:
        # There are only 5 clients actually in the current firmware
        # but we in case it will be ever changed in the future
        for num in range(1, 10):
            # Check if this client exists
            if f"client{num}_state" not in vpnmap:
                break

            # Get client data
            # We define default state as 0, since it's not always present
            client_state = AsusOVPNClient(
                safe_int(vpnmap.get(f"client{num}_state"), default=0)
            )
            client_errno = safe_int(vpnmap.get(f"client{num}_errno"))

            # Assign client data
            vpn["client"][num] = {
                "state": client_state,
                "errno": client_errno,
            }

        # Server data. Usually only 2 servers
        # but we in case it will be ever changed in the future
        for num in range(1, 5):
            # Check if this server exists
            if f"server{num}_state" not in vpnmap:
                break

            # Get server data
            # We define default state as 0, since it's not always present
            server_state = AsusOVPNServer(
                safe_int(vpnmap.get(f"server{num}_state"), default=0)
            )

            # Assign server data
            vpn["server"][num] = {
                "state": server_state,
            }

    return vpn



"""Home Assistant module for AsusRouter.

This module is used to convert some of the AsusRouter data
to a format easy to handle by Home Assistant integration.

Native AsusRouter integration: https://github.com/vaskivskyi/ha-asusrouter."""


from __future__ import annotations

import logging
from enum import Enum
from typing import Any, Optional

from asusrouter.modules.data import AsusData
from asusrouter.modules.state import AsusState
from asusrouter.modules.vpnc import AsusVPNC
from asusrouter.tools.converters import flatten_dict, list_from_dict

_LOGGER = logging.getLogger(__name__)

SENSORS_CPU = ["total", "usage", "used"]
SENSORS_NETWORK = ["rx", "rx_speed", "tx", "tx_speed"]
SENSORS_VPN = {
    "client": [
        "state",
        "remote",
        "datetime",
        "tun_tap_read",
        "tun_tap_write",
        "tcp_udp_read",
        "tcp_udp_write",
        "auth_read",
        "pre_compress",
        "post_compress",
        "pre_decompress",
        "post_decompress",
    ],
    "server": ["state", "client_list", "routing_table"],
}


def convert_to_ha_sensors(data: dict[str, Any], datatype: AsusData) -> list[str]:
    """Convert available data to the list of sensors
    compatible with Home Assistant."""

    sensors = []

    match datatype:
        case AsusData.CPU:
            sensors = convert_to_ha_sensors_by_map(data, SENSORS_CPU)
        case AsusData.NETWORK:
            sensors = convert_to_ha_sensors_by_map(data, SENSORS_NETWORK)
        case AsusData.OPENVPN:
            sensors = convert_to_ha_sensors_by_map_2(data, SENSORS_VPN)
        # case AsusData.PORTS:
        case _:
            sensors = convert_to_ha_sensors_list(data)

    return sensors


def convert_to_ha_data(data: dict[str, Any]) -> dict[str, Any]:
    """Convert available data to the HA-compatible dictionary."""

    def convert_recursive(data: dict[str, Any]) -> dict[str, Any]:
        """Convert data to the HA-compatible dictionary recursively."""
        return {
            key: convert_recursive(value)
            if isinstance(value, dict)
            else convert_to_ha_state_bool(value)
            if key.endswith("state") or key.endswith("link")
            else value
            for key, value in data.items()
        }

    # Flatten the dictionary
    # Skip all the `list`, `clients` etc keys - this data should be preserved
    output = flatten_dict(data, exclude=["list", "clients", "rules"])

    # Convert values to HA-compatible format
    if output is not None:
        output = convert_recursive(output)
        return output

    return {}


def convert_to_ha_sensors_by_map(
    data: dict[str, Any], sensor_map: list[str]
) -> list[str]:
    """Convert available data to the list of sensors
    using static map."""

    _LOGGER.debug("Converting data to the list of sensors by map: %s", data)

    sensors = []

    for sensor in data:
        for sensor_type in sensor_map:
            sensors.append(f"{sensor}_{sensor_type}")

    return sensors


def convert_to_ha_sensors_by_map_2(
    data: dict[str, Any], sensor_map: dict[str, list[str]]
) -> list[str]:
    """Convert available data to the list of sensors
    using first two levels of the data and static map."""

    _LOGGER.debug("Converting data to the list of sensors by 2 levels: %s", data)

    sensors = []

    for sensor in data:
        for sensor_id in data[sensor]:
            for sensor_type in sensor_map.get(sensor, []):
                sensors.append(f"{sensor}_{sensor_id}_{sensor_type}")

    return sensors


def convert_to_ha_sensors_group(data: dict[str, Any]) -> list[str]:
    """Convert the top level of data to the list of sensors."""

    return list_from_dict(data)


def convert_to_ha_sensors_list(data: dict[str, Any]) -> list[str]:
    """Convert all the available data to the list of sensors."""

    return list_from_dict(convert_to_ha_data(data))


def convert_to_ha_state_bool(data: AsusState | Optional[bool]) -> Optional[bool]:
    """Convers native state to a binary state."""

    # Check whether the state is None
    if data is None:
        return None

    # Check whether the state is already a bool
    if isinstance(data, bool):
        return data

    # Special cases
    if isinstance(data, AsusVPNC):
        match data:
            case a if a in (AsusVPNC.CONNECTED, AsusVPNC.CONNECTING, AsusVPNC.ON):
                return True
            case a if a in (AsusVPNC.OFF, AsusVPNC.DISCONNECTED, AsusVPNC.ERROR):
                return False
            case _:
                return None

    # Check whether the state is based on (int, Enum)
    if isinstance(data, int) and isinstance(data, Enum):
        if data.value == 0:
            return False
        if data.value > 0:
            return True
        if data.value < 0:
            return None

    return None


def convert_to_ha_string(data: Any) -> str:
    """Converts data to a string."""

    # Check if we have None
    if data is None:
        return ""

    # If we have an enum
    # Check whether value is an enum or a string
    # If string, return it, if enum, go recursive
    if isinstance(data, Enum):
        return convert_to_ha_string(data.value)

    # Check if we have string
    if isinstance(data, str):
        return data

    # For any other type, return string representation
    return str(data)



"""Constants for hook endpoint module."""

# Network

from asusrouter.modules.connection import ConnectionState, ConnectionStatus
from asusrouter.modules.endpoint.wan import AsusDualWAN
from asusrouter.modules.ip_address import read_dns_ip_address, read_ip_address_type
from asusrouter.modules.openvpn import AsusOVPNServer
from asusrouter.modules.wireguard import AsusWireGuardServer
from asusrouter.tools.converters import (
    safe_bool,
    safe_int,
    safe_list_csv,
    safe_list_from_string,
    safe_timestamp_to_utc,
)

MAP_NETWORK: dict[str, str] = {
    "INTERNET": "wan",
    "INTERNET1": "usb",
    "WIRED": "wired",
    "BRIDGE": "bridge",
    "WIRELESS0": "2ghz",
    "WIRELESS1": "5ghz",
    "WIRELESS2": "5ghz2",
    "WIRELESS3": "6ghz",
    "LACP1": "lacp1",
    "LACP2": "lacp2",
}

MAP_OVPN_SERVER_388 = (
    ("vpn_server_unit", "unit", safe_int),
    ("vpn_serverx_dns", "response_to_dns", safe_bool),
    ("vpn_server_port", "port", safe_int),
    ("vpn_server_tls_keysize", "tls_keysize", safe_int),
    ("vpn_server_if", "interface"),
    ("vpn_server_proto", "protocol"),
    ("vpn_server_pdns", "advertise_dns", safe_bool),
    ("vpn_server_cipher", "cipher"),
    ("vpn_server_digest", "digest"),
    ("vpn_server_comp", "compression"),
    ("vpn_server_igncrt", "password_only", safe_bool),
    ("vpn_server_crypt", "auth_mode"),
    ("vpn_server_hmac", "hmac"),
    ("vpn_server_sn", "subnet"),
    ("vpn_server_nm", "netmask"),
    ("vpn_server_dhcp", "dhcp", safe_bool),
    ("vpn_server_r1", "server_r1"),
    ("vpn_server_r2", "server_r2"),
    ("vpn_server_local", "address"),
    ("vpn_server_remote", "remote_address"),
    ("vpn_server_plan", "allow_lan", safe_bool),
    ("vpn_server_rgw", "allow_wan", safe_bool),
    ("vpn_server_reneg", "tls_reneg", safe_int),
    ("vpn_server_ccd", "client_specific_config", safe_bool),
    ("vpn_server_c2c", "client_to_client", safe_bool),
    ("vpn_server_ccd_excl", "allow_specific_clients", safe_bool),
    ("vpn_server_ccd_val", "specific_clients"),
    ("vpn_serverx_clientlist", "clients"),
    ("VPNServer_enable", "state", [safe_int, AsusOVPNServer]),
)

MAP_SPEEDTEST = (
    ("ookla_state", "state", safe_int),
    ("ookla_speedtest_get_history", "history"),
    ("ookla_speedtest_get_servers", "servers"),
    ("ookla_speedtest_get_result", "result"),
)

MAP_WAN = (
    ("get_wan_unit", "unit", safe_int),
    ("link_internet", "link", [safe_int, ConnectionStatus]),
    ("link_wan", "link_0", [safe_int, ConnectionState]),
    ("link_wan1", "link_1", [safe_int, ConnectionState]),
    ("wans_mode", "dualwan_mode", AsusDualWAN),
    ("wans_dualwan", "dualwan_priority", safe_list_from_string),
    ("bond_wan", "aggregation_state", safe_bool),
    ("wanports_bond", "aggregation_ports", safe_list_from_string),
)

MAP_WAN_ITEM = (
    ("auxstate_t", "auxstate", [safe_int, ConnectionStatus]),
    ("primary", "primary", safe_bool),
    ("proto", "protocol", read_ip_address_type),
    ("realip_ip", "real_ip"),
    ("realip_state", "real_ip_state", safe_bool),
    ("state_t", "state", [safe_int, ConnectionStatus]),
    ("sbstate_t", "bstate", [safe_int, ConnectionStatus]),
)

MAP_WAN_ITEM_X = (
    ("dns", "dns", read_dns_ip_address),
    ("expires", "expires", safe_int),
    ("gateway", "gateway"),
    ("ipaddr", "ip_address"),
    ("lease", "lease", safe_int),
    ("netmask", "mask"),
)

MAP_WIREGUARD_SERVER = (
    ("wgs_enable", "state", [safe_int, AsusWireGuardServer]),
    ("wgs_lanaccess", "lan_access", safe_bool),
    ("wgs_addr", "address"),
    ("wgs_port", "port", safe_int),
    ("wgs_dns", "dns", safe_bool),
    ("wgs_nat6", "nat6", safe_bool),
    ("wgs_psk", "psk", safe_bool),
    ("wgs_alive", "keep_alive", safe_int),
    ("wgs_priv", "private_key"),
    ("wgs_pub", "public_key"),
    ("get_wgsc_status", "status"),
)

MAP_WIREGUARD_CLIENT = (
    ("enable", "enabled", safe_bool),
    ("name", "name"),
    ("addr", "address"),
    ("aips", "allowed_ips", safe_list_csv),
    ("caips", "client_allowed_ips", safe_list_csv),
)

MAP_VPNC_WIREGUARD = (
    ("enable", "state", safe_bool),
    ("nat", "nat", safe_bool),
    ("priv", "private_key"),
    ("addr", "address"),
    ("dns", "dns", safe_bool),
    ("mtu", "mtu", safe_int),
    ("ppub", "public_key"),
    ("psk", "psk"),
    ("aips", "allowed_ips", safe_list_csv),
    ("ep_addr", "endpoint_address"),
    ("ep_port", "endpoint_port", safe_int),
    ("alive", "keep_alive", safe_int),
)



"""Service module."""

from __future__ import annotations

import logging
from typing import Any, Awaitable, Callable, Optional, Tuple

from asusrouter.error import AsusRouterServiceError
from asusrouter.tools.converters import safe_bool, safe_int

_LOGGER = logging.getLogger(__name__)


async def async_call_service(
    callback: Callable[..., Awaitable[dict[str, Any]]],
    service: Optional[str],
    arguments: Optional[dict[str, Any]] = None,
    apply: bool = False,
    expect_modify: bool = True,
) -> Tuple[bool, Optional[int], Optional[int]]:
    """Call a service."""

    # Generate commands
    commands = {}

    # Service call provided
    if service is not None:
        commands["rc_service"] = service

    # Check arguments
    if not arguments:
        arguments = {}

    # Add apply command if requested
    if apply:
        arguments["action_mode"] = "apply"

    # Add arguments to the commands
    commands.update(arguments)

    # Send the commands
    try:
        result = await callback(commands)
    except Exception as ex:  # pylint: disable=broad-except
        raise ex

    if service is not None:
        # Check if the service is run
        run_service = result.get("run_service", None)
        if run_service != service:
            raise AsusRouterServiceError(f"Service not run. Raw result: {result}")

    _LOGGER.debug(
        "Service `%s` run with arguments `%s`. Result: `%s`", service, arguments, result
    )

    last_id = result.get("id") or arguments.get("id")
    last_id = safe_int(last_id)

    needed_time = safe_int(result.get("restart_needed_time"))
    # For all the services with setting ID we better wait
    # before we will get actual state change
    if needed_time is None and last_id is not None:
        needed_time = 5

    # Special services that won't return any result
    if arguments.get("action_mode") == "update_client_list":
        return (True, needed_time, last_id)

    if expect_modify:
        return (safe_bool(result.get("modify")) or False, needed_time, last_id)

    return (True, needed_time, last_id)



"""Port forwarding module."""

import logging
from dataclasses import dataclass
from enum import IntEnum
from typing import Any, Awaitable, Callable

_LOGGER = logging.getLogger(__name__)

KEY_PORT_FORWARDING_LIST = "vts_rulelist"
KEY_PORT_FORWARDING_STATE = "vts_enable_x"


@dataclass(frozen=True)
class PortForwardingRule:
    """Port forwarding class"""

    name: str = str()
    ip_address: str = str()
    port: str = str()
    protocol: str = str()
    ip_external: str = str()
    port_external: str = str()


class AsusPortForwarding(IntEnum):
    """Asus port forwarding state."""

    UNKNOWN = -999
    OFF = 0
    ON = 1


async def set_state(
    callback: Callable[..., Awaitable[bool]],
    state: AsusPortForwarding,
    **kwargs: Any,
) -> bool:
    """Set the parental control state."""

    # Check if state is available and valid
    if not isinstance(state, AsusPortForwarding) or not state.value in (0, 1):
        _LOGGER.debug("No state found in arguments")
        return False

    arguments = {KEY_PORT_FORWARDING_STATE: 1 if state == AsusPortForwarding.ON else 0}

    # Get the correct service call
    service = "restart_firewall"

    # Call the service
    return await callback(
        service=service,
        arguments=arguments,
        apply=True,
        expect_modify=kwargs.get("expect_modify", False),
    )



"""Port module for AsusRouter.

This module is for physical ports on the router, e.g. LAN, WAN, USB, etc."""

from enum import IntEnum, StrEnum


class PortSpeed(IntEnum):
    """Port speed class."""

    LINK_DOWN = 0
    LINK_10 = 10
    LINK_100 = 100
    LINK_1000 = 1000
    LINK_2500 = 2500
    LINK_5000 = 5000
    LINK_10000 = 10000

    UNKNOWN = -999


class USBSpeed(IntEnum):
    """USB speed class."""

    USB_DOWN = 0
    USB_20 = 480
    USB_30 = 5000
    USB_31 = 10000
    USB_32 = 20000

    UNKNOWN = -999


PORT_SPEED = {
    "t": PortSpeed.LINK_10,
    "X": PortSpeed.LINK_DOWN,
    "M": PortSpeed.LINK_100,
    "G": PortSpeed.LINK_1000,
    "Q": PortSpeed.LINK_2500,
    "F": PortSpeed.LINK_5000,
    "T": PortSpeed.LINK_10000,
}

PORT_SUBTYPE = {
    1: PORT_SPEED["M"],
    2: PORT_SPEED["M"],
    3: PORT_SPEED["G"],
    4: PORT_SPEED["F"],
    5: PORT_SPEED["Q"],
    6: PORT_SPEED["T"],
    7: PORT_SPEED["T"],
}


class PortType(StrEnum):
    """Port type class."""

    ETHERNET = "ethernet"
    LAN = "lan"
    MOCA = "moca"
    POWERLINE = "powerline"
    SFPP = "sfpp"
    USB = "usb"
    WAN = "wan"

    UNKNOWN = "unknown"


class PortCapability(IntEnum):
    """Port capability class."""

    WAN = 0
    LAN = 1
    GAME = 2
    PLC = 3
    WAN2 = 4
    WAN3 = 5
    SFPP = 6
    USB = 7
    MOBILE = 8
    WANLAN = 9
    MOCA = 10
    IPTV_BRIDGE = 26
    IPTV_VOIP = 27
    IPTV_STB = 28
    DUALWAN_SECONDARY = 29
    DUALWAN_PRIMARY = 30


# Map to convert port capabilities to port types
# The order is important!
PORT_CAP2TYPE = {
    PortCapability.WAN: PortType.WAN,
    PortCapability.LAN: PortType.LAN,
    PortCapability.USB: PortType.USB,
    PortCapability.MOCA: PortType.MOCA,
}

PORT_TYPE2LINK = {
    PortType.LAN: PortSpeed,
    PortType.MOCA: PortSpeed,
    PortType.USB: USBSpeed,
    PortType.WAN: PortSpeed,
}
